# @generated by generate_proto_mypy_stubs.py.  Do not edit!
import sys
from google.protobuf.descriptor import (
    Descriptor as google___protobuf___descriptor___Descriptor,
)

from google.protobuf.internal.containers import (
    RepeatedScalarFieldContainer as google___protobuf___internal___containers___RepeatedScalarFieldContainer,
)

from google.protobuf.message import (
    Message as google___protobuf___message___Message,
)

from typing import (
    Iterable as typing___Iterable,
    Mapping as typing___Mapping,
    MutableMapping as typing___MutableMapping,
    Optional as typing___Optional,
    Text as typing___Text,
)

from typing_extensions import (
    Literal as typing_extensions___Literal,
)


builtin___bool = bool
builtin___bytes = bytes
builtin___float = float
builtin___int = int


class ErrorProto(google___protobuf___message___Message):
    DESCRIPTOR: google___protobuf___descriptor___Descriptor = ...
    code = ... # type: builtin___int
    message = ... # type: typing___Text

    def __init__(self,
        *,
        code : typing___Optional[builtin___int] = None,
        message : typing___Optional[typing___Text] = None,
        ) -> None: ...
    @classmethod
    def FromString(cls, s: builtin___bytes) -> ErrorProto: ...
    def MergeFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
    def CopyFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
    if sys.version_info >= (3,):
        def ClearField(self, field_name: typing_extensions___Literal[u"code",u"message"]) -> None: ...
    else:
        def ClearField(self, field_name: typing_extensions___Literal[u"code",b"code",u"message",b"message"]) -> None: ...

class BranchRecord(google___protobuf___message___Message):
    DESCRIPTOR: google___protobuf___descriptor___Descriptor = ...
    name = ... # type: typing___Text
    commit = ... # type: typing___Text

    def __init__(self,
        *,
        name : typing___Optional[typing___Text] = None,
        commit : typing___Optional[typing___Text] = None,
        ) -> None: ...
    @classmethod
    def FromString(cls, s: builtin___bytes) -> BranchRecord: ...
    def MergeFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
    def CopyFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
    if sys.version_info >= (3,):
        def ClearField(self, field_name: typing_extensions___Literal[u"commit",u"name"]) -> None: ...
    else:
        def ClearField(self, field_name: typing_extensions___Literal[u"commit",b"commit",u"name",b"name"]) -> None: ...

class HashRecord(google___protobuf___message___Message):
    DESCRIPTOR: google___protobuf___descriptor___Descriptor = ...
    type = ... # type: typing___Text
    digest = ... # type: typing___Text

    def __init__(self,
        *,
        type : typing___Optional[typing___Text] = None,
        digest : typing___Optional[typing___Text] = None,
        ) -> None: ...
    @classmethod
    def FromString(cls, s: builtin___bytes) -> HashRecord: ...
    def MergeFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
    def CopyFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
    if sys.version_info >= (3,):
        def ClearField(self, field_name: typing_extensions___Literal[u"digest",u"type"]) -> None: ...
    else:
        def ClearField(self, field_name: typing_extensions___Literal[u"digest",b"digest",u"type",b"type"]) -> None: ...

class CommitRecord(google___protobuf___message___Message):
    DESCRIPTOR: google___protobuf___descriptor___Descriptor = ...
    parent = ... # type: builtin___bytes
    ref = ... # type: builtin___bytes
    spec = ... # type: builtin___bytes

    def __init__(self,
        *,
        parent : typing___Optional[builtin___bytes] = None,
        ref : typing___Optional[builtin___bytes] = None,
        spec : typing___Optional[builtin___bytes] = None,
        ) -> None: ...
    @classmethod
    def FromString(cls, s: builtin___bytes) -> CommitRecord: ...
    def MergeFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
    def CopyFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
    if sys.version_info >= (3,):
        def ClearField(self, field_name: typing_extensions___Literal[u"parent",u"ref",u"spec"]) -> None: ...
    else:
        def ClearField(self, field_name: typing_extensions___Literal[u"parent",b"parent",u"ref",b"ref",u"spec",b"spec"]) -> None: ...

class SchemaRecord(google___protobuf___message___Message):
    DESCRIPTOR: google___protobuf___descriptor___Descriptor = ...
    digest = ... # type: typing___Text
    blob = ... # type: builtin___bytes

    def __init__(self,
        *,
        digest : typing___Optional[typing___Text] = None,
        blob : typing___Optional[builtin___bytes] = None,
        ) -> None: ...
    @classmethod
    def FromString(cls, s: builtin___bytes) -> SchemaRecord: ...
    def MergeFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
    def CopyFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
    if sys.version_info >= (3,):
        def ClearField(self, field_name: typing_extensions___Literal[u"blob",u"digest"]) -> None: ...
    else:
        def ClearField(self, field_name: typing_extensions___Literal[u"blob",b"blob",u"digest",b"digest"]) -> None: ...

class PingRequest(google___protobuf___message___Message):
    DESCRIPTOR: google___protobuf___descriptor___Descriptor = ...

    def __init__(self,
        ) -> None: ...
    @classmethod
    def FromString(cls, s: builtin___bytes) -> PingRequest: ...
    def MergeFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
    def CopyFrom(self, other_msg: google___protobuf___message___Message) -> None: ...

class PingReply(google___protobuf___message___Message):
    DESCRIPTOR: google___protobuf___descriptor___Descriptor = ...
    result = ... # type: typing___Text

    def __init__(self,
        *,
        result : typing___Optional[typing___Text] = None,
        ) -> None: ...
    @classmethod
    def FromString(cls, s: builtin___bytes) -> PingReply: ...
    def MergeFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
    def CopyFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
    if sys.version_info >= (3,):
        def ClearField(self, field_name: typing_extensions___Literal[u"result"]) -> None: ...
    else:
        def ClearField(self, field_name: typing_extensions___Literal[u"result",b"result"]) -> None: ...

class GetClientConfigRequest(google___protobuf___message___Message):
    DESCRIPTOR: google___protobuf___descriptor___Descriptor = ...

    def __init__(self,
        ) -> None: ...
    @classmethod
    def FromString(cls, s: builtin___bytes) -> GetClientConfigRequest: ...
    def MergeFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
    def CopyFrom(self, other_msg: google___protobuf___message___Message) -> None: ...

class GetClientConfigReply(google___protobuf___message___Message):
    DESCRIPTOR: google___protobuf___descriptor___Descriptor = ...
    class ConfigEntry(google___protobuf___message___Message):
        DESCRIPTOR: google___protobuf___descriptor___Descriptor = ...
        key = ... # type: typing___Text
        value = ... # type: typing___Text

        def __init__(self,
            *,
            key : typing___Optional[typing___Text] = None,
            value : typing___Optional[typing___Text] = None,
            ) -> None: ...
        @classmethod
        def FromString(cls, s: builtin___bytes) -> GetClientConfigReply.ConfigEntry: ...
        def MergeFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
        def CopyFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
        if sys.version_info >= (3,):
            def ClearField(self, field_name: typing_extensions___Literal[u"key",u"value"]) -> None: ...
        else:
            def ClearField(self, field_name: typing_extensions___Literal[u"key",b"key",u"value",b"value"]) -> None: ...


    @property
    def config(self) -> typing___MutableMapping[typing___Text, typing___Text]: ...

    @property
    def error(self) -> ErrorProto: ...

    def __init__(self,
        *,
        config : typing___Optional[typing___Mapping[typing___Text, typing___Text]] = None,
        error : typing___Optional[ErrorProto] = None,
        ) -> None: ...
    @classmethod
    def FromString(cls, s: builtin___bytes) -> GetClientConfigReply: ...
    def MergeFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
    def CopyFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
    if sys.version_info >= (3,):
        def HasField(self, field_name: typing_extensions___Literal[u"error"]) -> builtin___bool: ...
        def ClearField(self, field_name: typing_extensions___Literal[u"config",u"error"]) -> None: ...
    else:
        def HasField(self, field_name: typing_extensions___Literal[u"error",b"error"]) -> builtin___bool: ...
        def ClearField(self, field_name: typing_extensions___Literal[u"config",b"config",u"error",b"error"]) -> None: ...

class FetchBranchRecordRequest(google___protobuf___message___Message):
    DESCRIPTOR: google___protobuf___descriptor___Descriptor = ...

    @property
    def rec(self) -> BranchRecord: ...

    def __init__(self,
        *,
        rec : typing___Optional[BranchRecord] = None,
        ) -> None: ...
    @classmethod
    def FromString(cls, s: builtin___bytes) -> FetchBranchRecordRequest: ...
    def MergeFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
    def CopyFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
    if sys.version_info >= (3,):
        def HasField(self, field_name: typing_extensions___Literal[u"rec"]) -> builtin___bool: ...
        def ClearField(self, field_name: typing_extensions___Literal[u"rec"]) -> None: ...
    else:
        def HasField(self, field_name: typing_extensions___Literal[u"rec",b"rec"]) -> builtin___bool: ...
        def ClearField(self, field_name: typing_extensions___Literal[u"rec",b"rec"]) -> None: ...

class FetchBranchRecordReply(google___protobuf___message___Message):
    DESCRIPTOR: google___protobuf___descriptor___Descriptor = ...

    @property
    def rec(self) -> BranchRecord: ...

    @property
    def error(self) -> ErrorProto: ...

    def __init__(self,
        *,
        rec : typing___Optional[BranchRecord] = None,
        error : typing___Optional[ErrorProto] = None,
        ) -> None: ...
    @classmethod
    def FromString(cls, s: builtin___bytes) -> FetchBranchRecordReply: ...
    def MergeFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
    def CopyFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
    if sys.version_info >= (3,):
        def HasField(self, field_name: typing_extensions___Literal[u"error",u"rec"]) -> builtin___bool: ...
        def ClearField(self, field_name: typing_extensions___Literal[u"error",u"rec"]) -> None: ...
    else:
        def HasField(self, field_name: typing_extensions___Literal[u"error",b"error",u"rec",b"rec"]) -> builtin___bool: ...
        def ClearField(self, field_name: typing_extensions___Literal[u"error",b"error",u"rec",b"rec"]) -> None: ...

class FetchDataRequest(google___protobuf___message___Message):
    DESCRIPTOR: google___protobuf___descriptor___Descriptor = ...
    raw_data = ... # type: builtin___bytes
    comp_nbytes = ... # type: builtin___int
    uncomp_nbytes = ... # type: builtin___int

    @property
    def error(self) -> ErrorProto: ...

    def __init__(self,
        *,
        raw_data : typing___Optional[builtin___bytes] = None,
        comp_nbytes : typing___Optional[builtin___int] = None,
        uncomp_nbytes : typing___Optional[builtin___int] = None,
        error : typing___Optional[ErrorProto] = None,
        ) -> None: ...
    @classmethod
    def FromString(cls, s: builtin___bytes) -> FetchDataRequest: ...
    def MergeFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
    def CopyFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
    if sys.version_info >= (3,):
        def HasField(self, field_name: typing_extensions___Literal[u"error"]) -> builtin___bool: ...
        def ClearField(self, field_name: typing_extensions___Literal[u"comp_nbytes",u"error",u"raw_data",u"uncomp_nbytes"]) -> None: ...
    else:
        def HasField(self, field_name: typing_extensions___Literal[u"error",b"error"]) -> builtin___bool: ...
        def ClearField(self, field_name: typing_extensions___Literal[u"comp_nbytes",b"comp_nbytes",u"error",b"error",u"raw_data",b"raw_data",u"uncomp_nbytes",b"uncomp_nbytes"]) -> None: ...

class FetchDataReply(google___protobuf___message___Message):
    DESCRIPTOR: google___protobuf___descriptor___Descriptor = ...
    raw_data = ... # type: builtin___bytes
    comp_nbytes = ... # type: builtin___int
    uncomp_nbytes = ... # type: builtin___int

    @property
    def error(self) -> ErrorProto: ...

    def __init__(self,
        *,
        raw_data : typing___Optional[builtin___bytes] = None,
        comp_nbytes : typing___Optional[builtin___int] = None,
        uncomp_nbytes : typing___Optional[builtin___int] = None,
        error : typing___Optional[ErrorProto] = None,
        ) -> None: ...
    @classmethod
    def FromString(cls, s: builtin___bytes) -> FetchDataReply: ...
    def MergeFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
    def CopyFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
    if sys.version_info >= (3,):
        def HasField(self, field_name: typing_extensions___Literal[u"error"]) -> builtin___bool: ...
        def ClearField(self, field_name: typing_extensions___Literal[u"comp_nbytes",u"error",u"raw_data",u"uncomp_nbytes"]) -> None: ...
    else:
        def HasField(self, field_name: typing_extensions___Literal[u"error",b"error"]) -> builtin___bool: ...
        def ClearField(self, field_name: typing_extensions___Literal[u"comp_nbytes",b"comp_nbytes",u"error",b"error",u"raw_data",b"raw_data",u"uncomp_nbytes",b"uncomp_nbytes"]) -> None: ...

class FetchCommitRequest(google___protobuf___message___Message):
    DESCRIPTOR: google___protobuf___descriptor___Descriptor = ...
    commit = ... # type: typing___Text

    def __init__(self,
        *,
        commit : typing___Optional[typing___Text] = None,
        ) -> None: ...
    @classmethod
    def FromString(cls, s: builtin___bytes) -> FetchCommitRequest: ...
    def MergeFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
    def CopyFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
    if sys.version_info >= (3,):
        def ClearField(self, field_name: typing_extensions___Literal[u"commit"]) -> None: ...
    else:
        def ClearField(self, field_name: typing_extensions___Literal[u"commit",b"commit"]) -> None: ...

class FetchCommitReply(google___protobuf___message___Message):
    DESCRIPTOR: google___protobuf___descriptor___Descriptor = ...
    commit = ... # type: typing___Text
    total_byte_size = ... # type: builtin___int

    @property
    def record(self) -> CommitRecord: ...

    @property
    def error(self) -> ErrorProto: ...

    def __init__(self,
        *,
        commit : typing___Optional[typing___Text] = None,
        total_byte_size : typing___Optional[builtin___int] = None,
        record : typing___Optional[CommitRecord] = None,
        error : typing___Optional[ErrorProto] = None,
        ) -> None: ...
    @classmethod
    def FromString(cls, s: builtin___bytes) -> FetchCommitReply: ...
    def MergeFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
    def CopyFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
    if sys.version_info >= (3,):
        def HasField(self, field_name: typing_extensions___Literal[u"error",u"record"]) -> builtin___bool: ...
        def ClearField(self, field_name: typing_extensions___Literal[u"commit",u"error",u"record",u"total_byte_size"]) -> None: ...
    else:
        def HasField(self, field_name: typing_extensions___Literal[u"error",b"error",u"record",b"record"]) -> builtin___bool: ...
        def ClearField(self, field_name: typing_extensions___Literal[u"commit",b"commit",u"error",b"error",u"record",b"record",u"total_byte_size",b"total_byte_size"]) -> None: ...

class FetchSchemaRequest(google___protobuf___message___Message):
    DESCRIPTOR: google___protobuf___descriptor___Descriptor = ...

    @property
    def rec(self) -> SchemaRecord: ...

    def __init__(self,
        *,
        rec : typing___Optional[SchemaRecord] = None,
        ) -> None: ...
    @classmethod
    def FromString(cls, s: builtin___bytes) -> FetchSchemaRequest: ...
    def MergeFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
    def CopyFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
    if sys.version_info >= (3,):
        def HasField(self, field_name: typing_extensions___Literal[u"rec"]) -> builtin___bool: ...
        def ClearField(self, field_name: typing_extensions___Literal[u"rec"]) -> None: ...
    else:
        def HasField(self, field_name: typing_extensions___Literal[u"rec",b"rec"]) -> builtin___bool: ...
        def ClearField(self, field_name: typing_extensions___Literal[u"rec",b"rec"]) -> None: ...

class FetchSchemaReply(google___protobuf___message___Message):
    DESCRIPTOR: google___protobuf___descriptor___Descriptor = ...

    @property
    def rec(self) -> SchemaRecord: ...

    @property
    def error(self) -> ErrorProto: ...

    def __init__(self,
        *,
        rec : typing___Optional[SchemaRecord] = None,
        error : typing___Optional[ErrorProto] = None,
        ) -> None: ...
    @classmethod
    def FromString(cls, s: builtin___bytes) -> FetchSchemaReply: ...
    def MergeFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
    def CopyFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
    if sys.version_info >= (3,):
        def HasField(self, field_name: typing_extensions___Literal[u"error",u"rec"]) -> builtin___bool: ...
        def ClearField(self, field_name: typing_extensions___Literal[u"error",u"rec"]) -> None: ...
    else:
        def HasField(self, field_name: typing_extensions___Literal[u"error",b"error",u"rec",b"rec"]) -> builtin___bool: ...
        def ClearField(self, field_name: typing_extensions___Literal[u"error",b"error",u"rec",b"rec"]) -> None: ...

class PushBranchRecordRequest(google___protobuf___message___Message):
    DESCRIPTOR: google___protobuf___descriptor___Descriptor = ...

    @property
    def rec(self) -> BranchRecord: ...

    def __init__(self,
        *,
        rec : typing___Optional[BranchRecord] = None,
        ) -> None: ...
    @classmethod
    def FromString(cls, s: builtin___bytes) -> PushBranchRecordRequest: ...
    def MergeFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
    def CopyFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
    if sys.version_info >= (3,):
        def HasField(self, field_name: typing_extensions___Literal[u"rec"]) -> builtin___bool: ...
        def ClearField(self, field_name: typing_extensions___Literal[u"rec"]) -> None: ...
    else:
        def HasField(self, field_name: typing_extensions___Literal[u"rec",b"rec"]) -> builtin___bool: ...
        def ClearField(self, field_name: typing_extensions___Literal[u"rec",b"rec"]) -> None: ...

class PushBranchRecordReply(google___protobuf___message___Message):
    DESCRIPTOR: google___protobuf___descriptor___Descriptor = ...

    @property
    def error(self) -> ErrorProto: ...

    def __init__(self,
        *,
        error : typing___Optional[ErrorProto] = None,
        ) -> None: ...
    @classmethod
    def FromString(cls, s: builtin___bytes) -> PushBranchRecordReply: ...
    def MergeFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
    def CopyFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
    if sys.version_info >= (3,):
        def HasField(self, field_name: typing_extensions___Literal[u"error"]) -> builtin___bool: ...
        def ClearField(self, field_name: typing_extensions___Literal[u"error"]) -> None: ...
    else:
        def HasField(self, field_name: typing_extensions___Literal[u"error",b"error"]) -> builtin___bool: ...
        def ClearField(self, field_name: typing_extensions___Literal[u"error",b"error"]) -> None: ...

class PushDataRequest(google___protobuf___message___Message):
    DESCRIPTOR: google___protobuf___descriptor___Descriptor = ...
    raw_data = ... # type: builtin___bytes
    comp_nbytes = ... # type: builtin___int
    uncomp_nbytes = ... # type: builtin___int

    @property
    def error(self) -> ErrorProto: ...

    def __init__(self,
        *,
        raw_data : typing___Optional[builtin___bytes] = None,
        comp_nbytes : typing___Optional[builtin___int] = None,
        uncomp_nbytes : typing___Optional[builtin___int] = None,
        error : typing___Optional[ErrorProto] = None,
        ) -> None: ...
    @classmethod
    def FromString(cls, s: builtin___bytes) -> PushDataRequest: ...
    def MergeFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
    def CopyFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
    if sys.version_info >= (3,):
        def HasField(self, field_name: typing_extensions___Literal[u"error"]) -> builtin___bool: ...
        def ClearField(self, field_name: typing_extensions___Literal[u"comp_nbytes",u"error",u"raw_data",u"uncomp_nbytes"]) -> None: ...
    else:
        def HasField(self, field_name: typing_extensions___Literal[u"error",b"error"]) -> builtin___bool: ...
        def ClearField(self, field_name: typing_extensions___Literal[u"comp_nbytes",b"comp_nbytes",u"error",b"error",u"raw_data",b"raw_data",u"uncomp_nbytes",b"uncomp_nbytes"]) -> None: ...

class PushDataReply(google___protobuf___message___Message):
    DESCRIPTOR: google___protobuf___descriptor___Descriptor = ...

    @property
    def error(self) -> ErrorProto: ...

    def __init__(self,
        *,
        error : typing___Optional[ErrorProto] = None,
        ) -> None: ...
    @classmethod
    def FromString(cls, s: builtin___bytes) -> PushDataReply: ...
    def MergeFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
    def CopyFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
    if sys.version_info >= (3,):
        def HasField(self, field_name: typing_extensions___Literal[u"error"]) -> builtin___bool: ...
        def ClearField(self, field_name: typing_extensions___Literal[u"error"]) -> None: ...
    else:
        def HasField(self, field_name: typing_extensions___Literal[u"error",b"error"]) -> builtin___bool: ...
        def ClearField(self, field_name: typing_extensions___Literal[u"error",b"error"]) -> None: ...

class PushCommitRequest(google___protobuf___message___Message):
    DESCRIPTOR: google___protobuf___descriptor___Descriptor = ...
    commit = ... # type: typing___Text
    total_byte_size = ... # type: builtin___int

    @property
    def record(self) -> CommitRecord: ...

    def __init__(self,
        *,
        commit : typing___Optional[typing___Text] = None,
        total_byte_size : typing___Optional[builtin___int] = None,
        record : typing___Optional[CommitRecord] = None,
        ) -> None: ...
    @classmethod
    def FromString(cls, s: builtin___bytes) -> PushCommitRequest: ...
    def MergeFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
    def CopyFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
    if sys.version_info >= (3,):
        def HasField(self, field_name: typing_extensions___Literal[u"record"]) -> builtin___bool: ...
        def ClearField(self, field_name: typing_extensions___Literal[u"commit",u"record",u"total_byte_size"]) -> None: ...
    else:
        def HasField(self, field_name: typing_extensions___Literal[u"record",b"record"]) -> builtin___bool: ...
        def ClearField(self, field_name: typing_extensions___Literal[u"commit",b"commit",u"record",b"record",u"total_byte_size",b"total_byte_size"]) -> None: ...

class PushCommitReply(google___protobuf___message___Message):
    DESCRIPTOR: google___protobuf___descriptor___Descriptor = ...

    @property
    def error(self) -> ErrorProto: ...

    def __init__(self,
        *,
        error : typing___Optional[ErrorProto] = None,
        ) -> None: ...
    @classmethod
    def FromString(cls, s: builtin___bytes) -> PushCommitReply: ...
    def MergeFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
    def CopyFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
    if sys.version_info >= (3,):
        def HasField(self, field_name: typing_extensions___Literal[u"error"]) -> builtin___bool: ...
        def ClearField(self, field_name: typing_extensions___Literal[u"error"]) -> None: ...
    else:
        def HasField(self, field_name: typing_extensions___Literal[u"error",b"error"]) -> builtin___bool: ...
        def ClearField(self, field_name: typing_extensions___Literal[u"error",b"error"]) -> None: ...

class PushSchemaRequest(google___protobuf___message___Message):
    DESCRIPTOR: google___protobuf___descriptor___Descriptor = ...

    @property
    def rec(self) -> SchemaRecord: ...

    def __init__(self,
        *,
        rec : typing___Optional[SchemaRecord] = None,
        ) -> None: ...
    @classmethod
    def FromString(cls, s: builtin___bytes) -> PushSchemaRequest: ...
    def MergeFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
    def CopyFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
    if sys.version_info >= (3,):
        def HasField(self, field_name: typing_extensions___Literal[u"rec"]) -> builtin___bool: ...
        def ClearField(self, field_name: typing_extensions___Literal[u"rec"]) -> None: ...
    else:
        def HasField(self, field_name: typing_extensions___Literal[u"rec",b"rec"]) -> builtin___bool: ...
        def ClearField(self, field_name: typing_extensions___Literal[u"rec",b"rec"]) -> None: ...

class PushSchemaReply(google___protobuf___message___Message):
    DESCRIPTOR: google___protobuf___descriptor___Descriptor = ...

    @property
    def error(self) -> ErrorProto: ...

    def __init__(self,
        *,
        error : typing___Optional[ErrorProto] = None,
        ) -> None: ...
    @classmethod
    def FromString(cls, s: builtin___bytes) -> PushSchemaReply: ...
    def MergeFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
    def CopyFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
    if sys.version_info >= (3,):
        def HasField(self, field_name: typing_extensions___Literal[u"error"]) -> builtin___bool: ...
        def ClearField(self, field_name: typing_extensions___Literal[u"error"]) -> None: ...
    else:
        def HasField(self, field_name: typing_extensions___Literal[u"error",b"error"]) -> builtin___bool: ...
        def ClearField(self, field_name: typing_extensions___Literal[u"error",b"error"]) -> None: ...

class FindMissingCommitsRequest(google___protobuf___message___Message):
    DESCRIPTOR: google___protobuf___descriptor___Descriptor = ...
    commits = ... # type: google___protobuf___internal___containers___RepeatedScalarFieldContainer[typing___Text]

    @property
    def branch(self) -> BranchRecord: ...

    def __init__(self,
        *,
        commits : typing___Optional[typing___Iterable[typing___Text]] = None,
        branch : typing___Optional[BranchRecord] = None,
        ) -> None: ...
    @classmethod
    def FromString(cls, s: builtin___bytes) -> FindMissingCommitsRequest: ...
    def MergeFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
    def CopyFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
    if sys.version_info >= (3,):
        def HasField(self, field_name: typing_extensions___Literal[u"branch"]) -> builtin___bool: ...
        def ClearField(self, field_name: typing_extensions___Literal[u"branch",u"commits"]) -> None: ...
    else:
        def HasField(self, field_name: typing_extensions___Literal[u"branch",b"branch"]) -> builtin___bool: ...
        def ClearField(self, field_name: typing_extensions___Literal[u"branch",b"branch",u"commits",b"commits"]) -> None: ...

class FindMissingCommitsReply(google___protobuf___message___Message):
    DESCRIPTOR: google___protobuf___descriptor___Descriptor = ...
    commits = ... # type: google___protobuf___internal___containers___RepeatedScalarFieldContainer[typing___Text]

    @property
    def branch(self) -> BranchRecord: ...

    @property
    def error(self) -> ErrorProto: ...

    def __init__(self,
        *,
        commits : typing___Optional[typing___Iterable[typing___Text]] = None,
        branch : typing___Optional[BranchRecord] = None,
        error : typing___Optional[ErrorProto] = None,
        ) -> None: ...
    @classmethod
    def FromString(cls, s: builtin___bytes) -> FindMissingCommitsReply: ...
    def MergeFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
    def CopyFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
    if sys.version_info >= (3,):
        def HasField(self, field_name: typing_extensions___Literal[u"branch",u"error"]) -> builtin___bool: ...
        def ClearField(self, field_name: typing_extensions___Literal[u"branch",u"commits",u"error"]) -> None: ...
    else:
        def HasField(self, field_name: typing_extensions___Literal[u"branch",b"branch",u"error",b"error"]) -> builtin___bool: ...
        def ClearField(self, field_name: typing_extensions___Literal[u"branch",b"branch",u"commits",b"commits",u"error",b"error"]) -> None: ...

class FindMissingHashRecordsRequest(google___protobuf___message___Message):
    DESCRIPTOR: google___protobuf___descriptor___Descriptor = ...
    commit = ... # type: typing___Text
    hashs = ... # type: builtin___bytes
    total_byte_size = ... # type: builtin___int

    def __init__(self,
        *,
        commit : typing___Optional[typing___Text] = None,
        hashs : typing___Optional[builtin___bytes] = None,
        total_byte_size : typing___Optional[builtin___int] = None,
        ) -> None: ...
    @classmethod
    def FromString(cls, s: builtin___bytes) -> FindMissingHashRecordsRequest: ...
    def MergeFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
    def CopyFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
    if sys.version_info >= (3,):
        def ClearField(self, field_name: typing_extensions___Literal[u"commit",u"hashs",u"total_byte_size"]) -> None: ...
    else:
        def ClearField(self, field_name: typing_extensions___Literal[u"commit",b"commit",u"hashs",b"hashs",u"total_byte_size",b"total_byte_size"]) -> None: ...

class FindMissingHashRecordsReply(google___protobuf___message___Message):
    DESCRIPTOR: google___protobuf___descriptor___Descriptor = ...
    commit = ... # type: typing___Text
    hashs = ... # type: builtin___bytes
    total_byte_size = ... # type: builtin___int

    @property
    def error(self) -> ErrorProto: ...

    def __init__(self,
        *,
        commit : typing___Optional[typing___Text] = None,
        hashs : typing___Optional[builtin___bytes] = None,
        total_byte_size : typing___Optional[builtin___int] = None,
        error : typing___Optional[ErrorProto] = None,
        ) -> None: ...
    @classmethod
    def FromString(cls, s: builtin___bytes) -> FindMissingHashRecordsReply: ...
    def MergeFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
    def CopyFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
    if sys.version_info >= (3,):
        def HasField(self, field_name: typing_extensions___Literal[u"error"]) -> builtin___bool: ...
        def ClearField(self, field_name: typing_extensions___Literal[u"commit",u"error",u"hashs",u"total_byte_size"]) -> None: ...
    else:
        def HasField(self, field_name: typing_extensions___Literal[u"error",b"error"]) -> builtin___bool: ...
        def ClearField(self, field_name: typing_extensions___Literal[u"commit",b"commit",u"error",b"error",u"hashs",b"hashs",u"total_byte_size",b"total_byte_size"]) -> None: ...

class FindMissingSchemasRequest(google___protobuf___message___Message):
    DESCRIPTOR: google___protobuf___descriptor___Descriptor = ...
    commit = ... # type: typing___Text
    schema_digests = ... # type: google___protobuf___internal___containers___RepeatedScalarFieldContainer[typing___Text]

    def __init__(self,
        *,
        commit : typing___Optional[typing___Text] = None,
        schema_digests : typing___Optional[typing___Iterable[typing___Text]] = None,
        ) -> None: ...
    @classmethod
    def FromString(cls, s: builtin___bytes) -> FindMissingSchemasRequest: ...
    def MergeFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
    def CopyFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
    if sys.version_info >= (3,):
        def ClearField(self, field_name: typing_extensions___Literal[u"commit",u"schema_digests"]) -> None: ...
    else:
        def ClearField(self, field_name: typing_extensions___Literal[u"commit",b"commit",u"schema_digests",b"schema_digests"]) -> None: ...

class FindMissingSchemasReply(google___protobuf___message___Message):
    DESCRIPTOR: google___protobuf___descriptor___Descriptor = ...
    commit = ... # type: typing___Text
    schema_digests = ... # type: google___protobuf___internal___containers___RepeatedScalarFieldContainer[typing___Text]

    @property
    def error(self) -> ErrorProto: ...

    def __init__(self,
        *,
        commit : typing___Optional[typing___Text] = None,
        schema_digests : typing___Optional[typing___Iterable[typing___Text]] = None,
        error : typing___Optional[ErrorProto] = None,
        ) -> None: ...
    @classmethod
    def FromString(cls, s: builtin___bytes) -> FindMissingSchemasReply: ...
    def MergeFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
    def CopyFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
    if sys.version_info >= (3,):
        def HasField(self, field_name: typing_extensions___Literal[u"error"]) -> builtin___bool: ...
        def ClearField(self, field_name: typing_extensions___Literal[u"commit",u"error",u"schema_digests"]) -> None: ...
    else:
        def HasField(self, field_name: typing_extensions___Literal[u"error",b"error"]) -> builtin___bool: ...
        def ClearField(self, field_name: typing_extensions___Literal[u"commit",b"commit",u"error",b"error",u"schema_digests",b"schema_digests"]) -> None: ...



import pytest
import numpy as np


def create_meta_nt(name):
    from hangar.records import MetadataRecordKey
    res = MetadataRecordKey(name)
    return res


class TestReaderWriterDiff(object):

    @pytest.mark.parametrize('writer', [False, True])
    def test_diff_by_commit_and_branch(self, repo_2_br_no_conf, writer):
        repo = repo_2_br_no_conf
        testco = repo.checkout(branch='testbranch')
        masterco = repo.checkout(write=writer, branch='master')
        commit_diffs = masterco.diff.commit(testco.commit_hash)
        branch_diffs = masterco.diff.branch('testbranch')
        assert commit_diffs == branch_diffs
        testco.close()
        masterco.close()

    @pytest.mark.parametrize('writer', [False, True])
    def test_diff_with_wrong_commit_hash(self, repo_2_br_no_conf, writer):
        repo = repo_2_br_no_conf
        testco = repo.checkout(branch='testbranch')
        masterco = repo.checkout(write=writer, branch='master')
        wrong_commit_hash = testco.commit_hash + 'WrongHash'
        with pytest.raises(ValueError):
            masterco.diff.commit(wrong_commit_hash)
        testco.close()
        masterco.close()

    @pytest.mark.parametrize('writer', [False, True])
    def test_diff_with_wrong_branch_name(self, repo_1_br_no_conf, writer):
        repo = repo_1_br_no_conf
        masterco = repo.checkout(write=writer, branch='master')
        with pytest.raises(ValueError):
            masterco.diff.branch('wrong_branch_name')
        masterco.close()

    @pytest.mark.parametrize('writer', [False, True])
    def test_comparing_diffs_of_dev_and_master(self, repo_1_br_no_conf, writer):
        repo = repo_1_br_no_conf
        dummyData = np.arange(50)

        # mutating and removing data from testbranch
        testco = repo.checkout(write=True, branch='testbranch')
        testco.columns['dummy']['1'] = dummyData
        del testco.columns['dummy']['2']
        testco.commit("mutation and removal")
        testco.close()

        co1 = repo.checkout(write=writer, branch='master')
        diffdata1 = co1.diff.branch('testbranch')
        diffs1 = diffdata1.diff
        co1.close()

        co2 = repo.checkout(write=writer, branch='testbranch')
        diffdata2 = co2.diff.branch('master')
        diffs2 = diffdata2.diff
        co2.close()

        assert diffs1.added.samples == diffs2.added.samples
        assert diffs1.deleted.samples == diffs2.deleted.samples
        assert diffs1.mutated.samples == diffs2.mutated.samples

    @pytest.mark.parametrize('writer', [False, True])
    def test_diff_data_samples(self, repo_1_br_no_conf, writer):
        repo = repo_1_br_no_conf
        dummyData = np.arange(50)

        # mutating and removing data from testbranch
        testco = repo.checkout(write=True, branch='testbranch')
        testco.columns['dummy']['1'] = dummyData
        del testco.columns['dummy']['2']
        testco.commit("mutation and removal")
        testco.close()

        co = repo.checkout(write=writer, branch='master')
        diffdata = co.diff.branch('testbranch')
        conflicts = diffdata.conflict
        assert conflicts.conflict is False

        diffs = diffdata.diff

        # testing columns and metadata that has no change
        assert len(diffs.added.samples) == 20
        assert len(diffs.mutated.samples) == 1
        assert len(diffs.deleted.samples) == 1

        assert len(diffs.added.schema) == 0
        assert len(diffs.deleted.schema) == 0
        assert len(diffs.mutated.schema) == 0

        for datarecord in diffs.added.samples:
            assert 9 < int(datarecord.sample) < 20
        for mutated in diffs.mutated.samples:
            assert mutated.sample == '1'
        co.close()

    @pytest.mark.parametrize('writer', [False, True])
    def test_sample_addition_conflict(self, repo_1_br_no_conf, writer):
        # t1
        repo = repo_1_br_no_conf
        dummyData = np.arange(50)

        # adding data in master
        co = repo.checkout(write=True, branch='master')
        dummyData[:] = 123
        co.columns['dummy']['55'] = dummyData
        co.commit('Adding data in master')
        co.close()

        # adding data in testbranch
        co = repo.checkout(write=True, branch='testbranch')
        dummyData[:] = 234
        co.columns['dummy']['55'] = dummyData
        co.commit('adding data in testbranch')
        co.close()

        co = repo.checkout(write=writer, branch='master')
        conflicts = co.diff.branch('testbranch').conflict
        assert conflicts.conflict is True
        assert len(conflicts.t1.samples) == 1
        for k in conflicts.t1.samples:
            assert k.sample == '55'
        co.close()

    @pytest.mark.parametrize('writer', [False, True])
    def test_sample_removal_conflict(self, repo_1_br_no_conf, writer):
        # t21 and t22
        dummyData = np.arange(50)
        dummyData[:] = 123
        repo = repo_1_br_no_conf
        co = repo.checkout(write=True, branch='master')
        del co.columns['dummy']['6']
        co.columns['dummy']['7'] = dummyData
        co.commit('removal & mutation in master')
        co.close()

        co = repo.checkout(write=True, branch='testbranch')
        co.columns['dummy']['6'] = dummyData
        del co.columns['dummy']['7']
        co.commit('removal & mutation in dev')
        co.close()

        co = repo.checkout(write=writer, branch='master')
        conflicts = co.diff.branch('testbranch').conflict
        assert len(conflicts.t21.samples) == 1
        assert len(conflicts.t22.samples) == 1
        for k in conflicts.t21.samples:
            assert k.sample == '6'
        for k in conflicts.t22.samples:
            assert k.sample == '7'
        co.close()

    @pytest.mark.parametrize('writer', [False, True])
    def test_sample_mutation_conflict(self, repo_1_br_no_conf, writer):
        # t3
        dummyData = np.arange(50)
        dummyData[:] = 123
        repo = repo_1_br_no_conf
        co = repo.checkout(write=True, branch='master')
        co.columns['dummy']['7'] = dummyData
        co.commit('mutation in master')
        co.close()

        co = repo.checkout(write=True, branch='testbranch')
        dummyData[:] = 234
        co.columns['dummy']['7'] = dummyData
        co.commit('mutation in dev')
        co.close()

        co = repo.checkout(write=writer, branch='master')
        conflicts = co.diff.branch('testbranch').conflict
        assert len(conflicts.t3.samples) == 1
        for k in conflicts.t3.samples:
            assert k.sample == '7'
        co.close()

    @pytest.mark.parametrize('writer', [False, True])
    def test_aset_addition_conflict(self, aset_samples_initialized_repo, writer):
        # t1
        repo = aset_samples_initialized_repo

        repo.create_branch('testbranch')
        co = repo.checkout(write=True, branch='master')
        co.add_ndarray_column(name='testing_aset', shape=(5, 7), dtype=np.float64)
        co.commit('aset init in master')
        co.close()

        co = repo.checkout(write=True, branch='testbranch')
        co.add_ndarray_column(name='testing_aset', shape=(7, 7), dtype=np.float64)
        co.commit('aset init in dev')
        co.close()

        co = repo.checkout(write=writer, branch='master')
        conflicts = co.diff.branch('testbranch').conflict
        assert len(conflicts.t1.schema) == 1
        for k in conflicts.t1.schema:
            assert k.column == 'testing_aset'
        co.close()

    @pytest.mark.parametrize('writer', [False, True])
    def test_aset_removal_conflict(self, aset_samples_initialized_repo, writer):
        # t21 and t22
        repo = aset_samples_initialized_repo
        co = repo.checkout(write=True, branch='master')
        co.add_ndarray_column(name='testing_aset1', shape=(5, 7), dtype=np.float64)
        co.add_ndarray_column(name='testing_aset2', shape=(5, 7), dtype=np.float64)
        co.commit('added asets')
        co.close()
        repo.create_branch('testbranch')

        co = repo.checkout(write=True, branch='master')
        del co.columns['testing_aset1']
        del co.columns['testing_aset2']
        co.add_ndarray_column(name='testing_aset2', shape=(5, 7), dtype=np.float32)
        co.commit('mutation and removal from master')
        co.close()

        co = repo.checkout(write=True, branch='testbranch')
        del co.columns['testing_aset1']
        del co.columns['testing_aset2']
        co.add_ndarray_column(name='testing_aset1', shape=(5, 7), dtype=np.float32)
        co.commit('mutation and removal from dev')
        co.close()

        co = repo.checkout(write=writer, branch='master')
        conflicts = co.diff.branch('testbranch')[1]
        assert len(conflicts.t21.schema) == 1
        assert len(conflicts.t22.schema) == 1
        assert list(conflicts.t21.schema.keys())[0].column == 'testing_aset1'
        assert list(conflicts.t22.schema.keys())[0].column == 'testing_aset2'
        co.close()

    @pytest.mark.parametrize('writer', [False, True])
    def test_aset_mutation_conflict(self, aset_samples_initialized_repo, writer):
        # t3
        repo = aset_samples_initialized_repo
        co = repo.checkout(write=True, branch='master')
        co.add_ndarray_column(name='testing_aset', shape=(5, 7), dtype=np.float64)
        co.commit('added aset')
        co.close()
        repo.create_branch('testbranch')

        co = repo.checkout(write=True, branch='master')
        del co.columns['testing_aset']
        co.add_ndarray_column(name='testing_aset', shape=(7, 7), dtype=np.float64)
        co.commit('mutation from master')
        co.close()

        co = repo.checkout(write=True, branch='testbranch')
        del co.columns['testing_aset']
        co.add_ndarray_column(name='testing_aset', shape=(5, 7), dtype=np.float32)
        co.commit('mutation from dev')
        co.close()

        co = repo.checkout(write=writer, branch='master')
        conflicts = co.diff.branch('testbranch')[1]
        assert len(conflicts.t3.schema) == 1
        assert list(conflicts.t3.schema.keys())[0].column == 'testing_aset'
        co.close()


    @pytest.mark.parametrize('writer', [False, True])
    def test_commits_inside_cm(self, aset_samples_initialized_repo, array5by7, writer):
        repo = aset_samples_initialized_repo
        repo.create_branch('testbranch')
        co = repo.checkout(write=True, branch='testbranch')
        aset = co.columns['writtenaset']
        aset2 = co.add_ndarray_column('aset2', prototype=array5by7)
        aset2[1] = array5by7
        with aset:
            aset[100] = array5by7
            co.commit('inside cm')
            aset[101] = array5by7
            co.commit('another commit inside cm')
        co.close()
        co = repo.checkout(write=writer, branch='testbranch')
        assert np.allclose(co.columns['writtenaset'][101], array5by7)
        diff = co.diff.branch('master').diff
        assert 'aset2' in [x.column for x in diff.added.schema.keys()]
        calledWithAset = False
        for record in diff.added.samples:
            if record.column == 'writtenaset':
                calledWithAset = True
                assert record.sample in [100, 101]
        assert calledWithAset is True
        co.close()


class TestWriterDiff(object):

    def test_status_and_staged_column(self, aset_samples_initialized_repo):
        repo = aset_samples_initialized_repo
        co = repo.checkout(write=True)
        co.add_str_column('DOESNOTEXIST')
        co['DOESNOTEXIST'][1] = 'foo'
        assert co.diff.status() == 'DIRTY'
        co.commit('init metadata')
        assert co.diff.status() == 'CLEAN'
        co.close()

    def test_status_and_staged_samples(self, aset_samples_initialized_repo):
        dummyData = np.zeros((5, 7))
        repo = aset_samples_initialized_repo
        co = repo.checkout()
        with pytest.raises(AttributeError):
            co.diff.status()  # Read checkout doesn't have status()

        co = repo.checkout(write=True)
        co.columns['writtenaset']['45'] = dummyData
        assert co.diff.status() == 'DIRTY'
        diff = co.diff.staged()
        calledWithAset = False
        for record in diff.diff.added.samples:
            if record.column == 'writtenaset':
                calledWithAset = True
                assert record.sample in '45'
        assert calledWithAset is True
        co.commit('adding')
        assert co.diff.status() == 'CLEAN'
        co.close()

    def test_status_and_staged_aset(self, aset_samples_initialized_repo):
        repo = aset_samples_initialized_repo
        co = repo.checkout(write=True)
        co.add_ndarray_column(name='sampleaset', shape=(3, 5), dtype=np.float32)
        assert co.diff.status() == 'DIRTY'
        diff = co.diff.staged()
        assert 'sampleaset' in [x.column for x in diff.diff.added.schema]
        co.commit('init aset')
        assert co.diff.status() == 'CLEAN'
        co.close()


def test_repo_diff_method_branch_names(aset_samples_initialized_repo):
    # t3
    repo = aset_samples_initialized_repo
    co = repo.checkout(write=True, branch='master')
    co.add_ndarray_column(name='testing_aset', shape=(5, 7), dtype=np.float64)
    co.commit('added aset')
    co.close()
    repo.create_branch('testbranch')

    co = repo.checkout(write=True, branch='master')
    del co.columns['testing_aset']
    co.add_ndarray_column(name='testing_aset', shape=(7, 7), dtype=np.float64)
    masterHEAD = co.commit('mutation from master')
    co.close()

    co = repo.checkout(write=True, branch='testbranch')
    del co.columns['testing_aset']
    co.add_ndarray_column(name='testing_aset', shape=(5, 7), dtype=np.float32)
    devHEAD = co.commit('mutation from dev')
    co.close()

    co = repo.checkout(write=False, branch='master')
    co_diff = co.diff.branch('testbranch')
    co.close()

    repo_diff = repo.diff('master', 'testbranch')
    assert co_diff == repo_diff


def test_repo_diff_method_commit_digests(aset_samples_initialized_repo):
    # t3
    repo = aset_samples_initialized_repo
    co = repo.checkout(write=True, branch='master')
    co.add_ndarray_column(name='testing_aset', shape=(5, 7), dtype=np.float64)
    co.commit('added aset')
    co.close()
    repo.create_branch('testbranch')

    co = repo.checkout(write=True, branch='master')
    del co.columns['testing_aset']
    co.add_ndarray_column(name='testing_aset', shape=(7, 7), dtype=np.float64)
    masterHEAD = co.commit('mutation from master')
    co.close()

    co = repo.checkout(write=True, branch='testbranch')
    del co.columns['testing_aset']
    co.add_ndarray_column(name='testing_aset', shape=(5, 7), dtype=np.float32)
    devHEAD = co.commit('mutation from dev')
    co.close()

    co = repo.checkout(write=False, branch='master')
    co_diff = co.diff.commit(devHEAD)
    co.close()

    repo_diff = repo.diff(masterHEAD, devHEAD)
    assert co_diff == repo_diff




def test_repo_diff_method_one_branch_one_commit_digest(aset_samples_initialized_repo):
    # t3
    repo = aset_samples_initialized_repo
    co = repo.checkout(write=True, branch='master')
    co.add_ndarray_column(name='testing_aset', shape=(5, 7), dtype=np.float64)
    co.commit('added aset')
    co.close()
    repo.create_branch('testbranch')

    co = repo.checkout(write=True, branch='master')
    del co.columns['testing_aset']
    co.add_ndarray_column(name='testing_aset', shape=(7, 7), dtype=np.float64)
    masterHEAD = co.commit('mutation from master')
    co.close()

    co = repo.checkout(write=True, branch='testbranch')
    del co.columns['testing_aset']
    co.add_ndarray_column(name='testing_aset', shape=(5, 7), dtype=np.float32)
    devHEAD = co.commit('mutation from dev')
    co.close()

    co = repo.checkout(write=False, branch='master')
    co_diff = co.diff.commit(devHEAD)
    co.close()

    repo_diff1 = repo.diff('master', devHEAD)
    assert co_diff == repo_diff1

    repo_diff2 = repo.diff(masterHEAD, 'testbranch')
    assert co_diff == repo_diff2



# memory efficient container classes for data backends specs.
# Allow for attribute access similar to named tuples.

cdef class HDF5_01_DataHashSpec:

    def __init__(self, str backend, str uid, str checksum, str dataset,
                 int dataset_idx, tuple shape):

        self.backend = backend
        self.uid = uid
        self.checksum = checksum
        self.dataset = dataset
        self.dataset_idx = dataset_idx
        self.shape = shape

    def __repr__(self):
        return (f'{self.__class__.__name__}('
                f'backend="{self.backend}", '
                f'uid="{self.uid}", '
                f'checksum="{self.checksum}", '
                f'dataset="{self.dataset}", '
                f'dataset_idx={self.dataset_idx}, '
                f'shape={self.shape})')

    def __iter__(self):
        for attr in ['backend', 'uid', 'checksum', 'dataset', 'dataset_idx', 'shape']:
            yield getattr(self, attr)

    @property
    def islocal(self):
        return True


cdef class HDF5_00_DataHashSpec:

    def __init__(self, str backend, str uid, str checksum,
                 str dataset, int dataset_idx, tuple shape):

        self.backend = backend
        self.uid = uid
        self.checksum = checksum
        self.dataset = dataset
        self.dataset_idx = dataset_idx
        self.shape = shape

    def __repr__(self):
        return (f'{self.__class__.__name__}('
                f'backend="{self.backend}", '
                f'uid="{self.uid}", '
                f'checksum="{self.checksum}", '
                f'dataset="{self.dataset}", '
                f'dataset_idx={self.dataset_idx}, '
                f'shape={self.shape})')

    def __iter__(self):
        for attr in ['backend', 'uid', 'checksum', 'dataset', 'dataset_idx', 'shape']:
            yield getattr(self, attr)

    @property
    def islocal(self):
        return True


cdef class NUMPY_10_DataHashSpec:

    def __init__(self, str backend, str uid, str checksum,
                 int collection_idx, tuple shape):

        self.backend = backend
        self.uid = uid
        self.checksum = checksum
        self.collection_idx = collection_idx
        self.shape = shape

    def __repr__(self):
        return (f'{self.__class__.__name__}('
                f'backend="{self.backend}", '
                f'uid="{self.uid}", '
                f'checksum="{self.checksum}", '
                f'collection_idx={self.collection_idx}, '
                f'shape={self.shape})')

    def __iter__(self):
        for attr in ['backend', 'uid', 'checksum', 'collection_idx', 'shape']:
            yield getattr(self, attr)

    @property
    def islocal(self):
        return True


cdef class LMDB_30_DataHashSpec:

    def __init__(self, str backend, str uid, str row_idx, str checksum):

        self.backend = backend
        self.uid = uid
        self.row_idx = row_idx
        self.checksum = checksum


    def __repr__(self):
        return (f'{self.__class__.__name__}('
                f'backend="{self.backend}", '
                f'uid="{self.uid}", '
                f'row_idx={self.row_idx}, '
                f'checksum="{self.checksum}")')

    def __iter__(self):
        for attr in ['backend', 'uid', 'row_idx', 'checksum']:
            yield getattr(self, attr)

    @property
    def islocal(self):
        return True


cdef class REMOTE_50_DataHashSpec:

    def __init__(self, str backend, str schema_hash):

        self.backend = backend
        self.schema_hash = schema_hash

    def __repr__(self):
        return (f'{self.__class__.__name__}('
                f'backend="{self.backend}", '
                f'schema_hash="{self.schema_hash}")')

    def __iter__(self):
        for attr in ['backend', 'schema_hash']:
            yield getattr(self, attr)

    @property
    def islocal(self):
        return False



import pytest
import numpy as np


@pytest.fixture(params=['00', '01'])
def be_filehandle(request):
    if request.param == '00':
        from hangar.backends.hdf5_00 import HDF5_00_FileHandles
        return HDF5_00_FileHandles
    elif request.param == '01':
        from hangar.backends.hdf5_01 import HDF5_01_FileHandles
        return HDF5_01_FileHandles
    else:
        raise ValueError(f'request param "{request.param}" for backend code unknown.')


@pytest.mark.parametrize('clib,clibCode',
                         [('blosc:blosclz', 0), ('blosc:lz4', 1),
                          ('blosc:lz4hc', 2), ('blosc:zlib', 4),
                          ('blosc:zstd', 5)])
@pytest.mark.parametrize('clevel', [1, 4, 8])
@pytest.mark.parametrize('cshuffle,cshuffleCode', [(None, 0), ('byte', 1), ('bit', 2)])
@pytest.mark.parametrize('beCode', ['00', '01'])
def test_blosc_filter_opts_result_in_correct_dataset_args(
        be_filehandle, clib, clibCode, clevel, cshuffle, cshuffleCode, beCode):

    out = be_filehandle._dataset_opts(complib=clib,
                                      complevel=clevel,
                                      shuffle=cshuffle)
    expected = {
        'compression': 32001,
        'compression_opts': (0, 0, 0, 0, clevel, cshuffleCode, clibCode),
        'shuffle': False}
    assert out == expected


@pytest.mark.parametrize('cshuffle,cshuffleCode', [(None, False), ('byte', True)])
def test_lzf_filter_opts_result_in_correct_dataset_args(be_filehandle, cshuffle, cshuffleCode):
    out = be_filehandle._dataset_opts(complib='lzf',
                                      complevel=None,
                                      shuffle=cshuffle)
    expected = {
        'compression': 'lzf',
        'compression_opts': None,
        'shuffle': cshuffleCode}
    assert out == expected


@pytest.mark.parametrize('clevel', [1, 4, 8])
@pytest.mark.parametrize('cshuffle,cshuffleCode', [(None, False), ('byte', True)])
def test_gzip_filter_opts_result_in_correct_dataset_args(be_filehandle, clevel, cshuffle, cshuffleCode):
    out = be_filehandle._dataset_opts(complib='gzip',
                                      complevel=clevel,
                                      shuffle=cshuffle)
    expected = {
        'compression': 'gzip',
        'compression_opts': clevel,
        'shuffle': cshuffleCode}
    assert out == expected


# ------------------------- test actual compression ---------------------------


@pytest.mark.parametrize('clib,clibCode',
                         [('blosc:blosclz', 0), ('blosc:lz4', 1),
                          ('blosc:lz4hc', 2), ('blosc:zlib', 4),
                          ('blosc:zstd', 5)])
@pytest.mark.parametrize('clevel', [1, 4, 8])
@pytest.mark.parametrize('cshuffle,cshuffleCode', [(None, 0), ('byte', 1), ('bit', 2)])
@pytest.mark.parametrize('be_code', ['00', '01'])
def test_arrayset_init_with_various_blosc_opts(repo, array5by7, clib, clibCode, clevel, cshuffle, cshuffleCode, be_code):

    opts = {
        'shuffle': cshuffle,
        'complib': clib,
        'complevel': clevel,
    }
    wco = repo.checkout(write=True)
    aset = wco.add_ndarray_column('aset', prototype=array5by7, backend=be_code, backend_options=opts)
    assert aset.backend == be_code
    with aset as a:
        for i in range(10):
            a[i] = array5by7 + i

    wuid = aset._be_fs[be_code].w_uid
    plist = aset._be_fs[be_code].wFp[wuid]['/0'].id.get_create_plist()
    _, _, resopts, _ = plist.get_filter(0)
    res_clevel, res_cshuffle, res_clib = resopts[4:7]
    assert res_clevel == clevel
    assert res_clib == clibCode
    assert res_cshuffle == cshuffleCode
    wco.commit('hi')
    wco.close()


@pytest.mark.parametrize('cshuffle,cshuffleCode', [(False, False), (True, True)])
@pytest.mark.parametrize('be_code', ['00', '01'])
def test_arrayset_init_with_various_lzf_opts(repo, array5by7, cshuffle, cshuffleCode, be_code):

    opts = {
        'shuffle': cshuffle,
        'complib': 'lzf',
        'complevel': None,
    }
    wco = repo.checkout(write=True)
    aset = wco.add_ndarray_column('aset', prototype=array5by7, backend=be_code, backend_options=opts)
    assert aset.backend == be_code
    with aset as a:
        for i in range(10):
            a[i] = array5by7 + i

    res_compression = aset._be_fs[be_code].wFp[aset._be_fs[be_code].w_uid]['/0'].compression
    res_shuffle = aset._be_fs[be_code].wFp[aset._be_fs[be_code].w_uid]['/0'].shuffle
    assert res_compression == 'lzf'
    assert res_shuffle == cshuffleCode
    wco.commit('hi')
    wco.close()


@pytest.mark.parametrize('clevel', [1, 4, 8])
@pytest.mark.parametrize('cshuffle,cshuffleCode', [(False, False), (True, True)])
@pytest.mark.parametrize('be_code', ['00', '01'])
def test_arrayset_init_with_various_gzip_opts(repo, array5by7, clevel, cshuffle, cshuffleCode, be_code):

    opts = {
        'shuffle': cshuffle,
        'complib': 'gzip',
        'complevel': clevel,
    }
    wco = repo.checkout(write=True)
    aset = wco.add_ndarray_column(
        'aset', prototype=array5by7, backend=be_code, backend_options=opts)
    assert aset.backend == be_code
    with aset as a:
        for i in range(10):
            a[i] = array5by7 + i

    res_compression = aset._be_fs[be_code].wFp[aset._be_fs[be_code].w_uid]['/0'].compression
    res_compression_opts = aset._be_fs[be_code].wFp[aset._be_fs[be_code].w_uid]['/0'].compression_opts
    res_shuffle = aset._be_fs[be_code].wFp[aset._be_fs[be_code].w_uid]['/0'].shuffle
    assert res_compression == 'gzip'
    assert res_shuffle == cshuffleCode
    assert res_compression_opts == clevel
    wco.commit('hi')
    wco.close()


@pytest.mark.parametrize('be_code', ['00', '01'])
def test_arrayset_overflows_collection_size_collection_count(be_code, repo, monkeypatch):
    if be_code == '00':
        from hangar.backends import hdf5_00
        monkeypatch.setattr(hdf5_00, 'COLLECTION_COUNT', 5)
        monkeypatch.setattr(hdf5_00, 'COLLECTION_SIZE', 10)
    elif be_code == '01':
        from hangar.backends import hdf5_01
        monkeypatch.setattr(hdf5_01, 'COLLECTION_COUNT', 5)
        monkeypatch.setattr(hdf5_01, 'COLLECTION_SIZE', 10)
    else:
        raise ValueError(f'be_code param "{be_code}" unknown.')

    wco = repo.checkout(write=True)
    proto = np.arange(50).astype(np.uint16)
    aset = wco.add_ndarray_column('aset', prototype=proto, backend=be_code)
    with aset as cm_aset:
        for i in range(500):
            proto[:] = i
            cm_aset[i] = proto
    assert aset._be_fs[be_code].hColsRemain == 4
    assert aset._be_fs[be_code].hMaxSize == 10
    wco.commit('hello')

    with aset as cm_aset:
        for i in range(500):
            proto[:] = i
            assert np.allclose(proto, cm_aset[i])
    wco.close()

    rco = repo.checkout()
    naset = rco.columns['aset']
    with naset as ncm_aset:
        for i in range(500):
            proto[:] = i
            assert np.allclose(proto, ncm_aset[i])
    rco.close()



import pytest

from hangar.optimized_utils import SizedDict


def test_sizeddict_maxsize_property():
    d = SizedDict(maxsize=5)
    assert d.maxsize == 5
    d2 = SizedDict(maxsize=10)
    assert d2.maxsize == 10


def test_sizeddict_setitem_no_overflow_retains_keys_and_len():
    d = SizedDict(maxsize=5)
    for i in range(5):
        d[i] = i

    assert len(d) == 5
    for i in range(5):
        assert i in d
        assert d[i] == i


def test_sizeddict_setitem_overflow_truncates_keys_and_len():
    d = SizedDict(maxsize=5)
    for i in range(10):
        d[i] = i

    assert len(d) == 5
    for i in range(0, 5):
        assert i not in d
        with pytest.raises(KeyError):
            _ = d[i]
    for i in range(5, 10):
        assert i in d
        assert d[i] == i


def test_sizeddict_update_no_overflow_retains_keys_and_len():
    d = SizedDict(maxsize=5)
    inp = {i: i for i in range(5)}
    d.update(inp)

    assert len(d) == 5
    for i in range(5):
        assert i in d
        assert d[i] == i


def test_sizeddict_updateoverflow_truncates_keys_and_len():
    d = SizedDict(maxsize=5)
    inp = {i: i for i in range(10)}
    d.update(inp)

    assert len(d) == 5
    for i in range(0, 5):
        assert i not in d
        with pytest.raises(KeyError):
            _ = d[i]
    for i in range(5, 10):
        assert i in d
        assert d[i] == i


def test_sizeddict_get_returns_default_on_missing_key():
    d = SizedDict()
    res = d.get('doesnotexist')
    assert res is None
    res = d.get('doesnotexist', default='foo')
    assert res == 'foo'


def test_sizeddict_delitem():
    d = SizedDict(maxsize=5)
    inp = {i: i for i in range(5)}
    d.update(inp)

    del d[3]
    assert len(d) == 4
    assert 3 not in d

    d['new'] = 'foo'
    assert len(d) == 5
    assert 'new' in d


def test_sizeddict_pop():
    d = SizedDict(maxsize=5)
    inp = {i: i for i in range(5)}
    d.update(inp)

    res = d.pop(0)
    assert res == 0
    assert len(d) == 4
    res = d.pop('doesnotexist', default='foo')
    assert res == 'foo'
    assert len(d) == 4


def test_sizeddict_popitem():
    d = SizedDict(maxsize=5)
    inp = {i: i for i in range(5)}
    d.update(inp)

    res = d.popitem()
    assert res == (4, 4)
    assert len(d) == 4
    res = d.popitem()
    assert res == (3, 3)
    assert len(d) == 3

    d['foo'] = 'bar'
    assert len(d) == 4
    res = d.popitem()
    assert res == ('foo', 'bar')
    assert len(d) == 3


def test_sizeddict_keys():
    d = SizedDict(maxsize=5)
    inp = {str(i): i for i in range(5)}
    d.update(inp)

    assert list(d.keys()) == list(inp.keys())
    for res_k, expected_k in zip(d.keys(), inp.keys()):
        assert res_k == expected_k


def test_sizeddict_values():
    d = SizedDict(maxsize=5)
    inp = {str(i): i for i in range(5)}
    d.update(inp)

    assert list(d.values()) == list(inp.values())
    for res_v, expected_v in zip(d.values(), inp.values()):
        assert res_v == expected_v


def test_sizeddict_keys():
    d = SizedDict(maxsize=5)
    inp = {str(i): i for i in range(5)}
    d.update(inp)

    assert list(d.items()) == list(inp.items())
    for res_kv, expected_kv in zip(d.items(), inp.items()):
        assert res_kv == expected_kv


def test_sizeddict_setdefault():
    d = SizedDict(maxsize=5)
    inp = {i: i for i in range(5)}
    d.update(inp)

    res = d.setdefault('doesnotexist')
    assert res is None
    assert len(d) == 5
    assert 'doesnotexist' in d
    assert d['doesnotexist'] is None

    res = d.setdefault('doesnotexist2', default=True)
    assert res is True
    assert len(d) == 5
    assert 'doesnotexist2' in d
    assert d['doesnotexist2'] is True

    res = d.setdefault(2, default=True)
    assert res == 2
    assert len(d) == 5
    assert 2 in d
    assert d[2] == 2


def test_sizeddict_clear():
    d = SizedDict(maxsize=5)
    inp = {i: i for i in range(5)}
    d.update(inp)

    assert len(d) == 5
    d.clear()
    assert len(d) == 0
    assert len(d._stack) == 0
    assert len(d._data) == 0
    assert d._stack_size == 0
    assert d._maxsize == 5


def test_sizeddict_repr():
    d = SizedDict(maxsize=5)
    inp = {i: i for i in range(5)}
    d.update(inp)

    expected = repr(inp)
    res = repr(d)
    assert res == expected


def test_sizeddict_is_pickleable():
    import pickle

    d = SizedDict(maxsize=5)
    inp = {i: i for i in range(5)}
    d.update(inp)

    pick = pickle.dumps(d, protocol=pickle.HIGHEST_PROTOCOL)
    res = pickle.loads(pick)

    assert res._maxsize == d._maxsize
    assert res._stack == d._stack
    assert res._stack_size == d._stack_size
    assert res._data == d._data



import pytest

import numpy as np
import time
from os.path import join as pjoin
from os import mkdir
from random import randint
import platform



@pytest.mark.parametrize('name', [
    'invalid\n', '\ninvalid', 'inv\name', 'inva/lid', 12, ' try', 'and this ',
    'VeryLongNameIsInvalidOver64CharactersNotAllowedVeryLongNameIsInva'])
def test_cannot_add_invalid_remote_names(repo, name):
    with pytest.raises(ValueError):
        repo.remote.add(name, 'localhost:50051')


def test_list_all_remotes_works(repo):

    remote_spec1 = repo.remote.add('origin', 'test')
    currentRemotes = repo.remote.list_all()

    assert len(currentRemotes) == 1
    currentSpec = currentRemotes[0]
    assert len(currentSpec) == 2
    assert currentSpec.name == 'origin'
    assert currentSpec.address == 'test'

    remote_spec2 = repo.remote.add('origin2', 'test2')
    currentRemotes = repo.remote.list_all()

    assert len(currentRemotes) == 2
    currentSpec = currentRemotes[0]
    assert currentSpec == remote_spec1
    assert len(currentSpec) == 2
    assert currentSpec.name == 'origin'
    assert currentSpec.address == 'test'
    currentSpec = currentRemotes[1]
    assert currentSpec == remote_spec2
    assert currentSpec.name == 'origin2'
    assert currentSpec.address == 'test2'


def test_cannot_add_remote_twice_with_same_name(repo):
    remote_spec = repo.remote.add('origin', 'test')
    assert remote_spec.name == 'origin'
    assert remote_spec.address == 'test'
    with pytest.raises(ValueError):
        repo.remote.add('origin', 'new')


def test_remote_remote_which_does_not_exist_fails(repo):
    with pytest.raises(ValueError):
        repo.remote.remove('origin')


def test_can_update_remote_after_removal(repo):
    remote_spec = repo.remote.add('origin', 'test')
    assert remote_spec.name == 'origin'
    assert remote_spec.address == 'test'
    channel_address_removed = repo.remote.remove('origin')
    assert channel_address_removed.name == 'origin'
    assert channel_address_removed.address == 'test'
    new_name = repo.remote.add('origin', 'test2')
    assert new_name.name == 'origin'
    assert new_name.address == 'test2'


def test_server_is_started_multiple_times_via_ping_pong(server_instance,
                                                        aset_samples_initialized_repo):
    # start multiple times and test that pings go through multiple times
    aset_samples_initialized_repo.remote.add('origin', server_instance)
    roundTripTime = aset_samples_initialized_repo.remote.ping('origin')
    assert isinstance(roundTripTime, float)


@pytest.mark.parametrize('nCommits,nSamples', [[1, 10], [5, 10]])
def test_push_and_clone_master_linear_history_multiple_commits(
        server_instance, repo, managed_tmpdir, array5by7, nCommits, nSamples):
    from hangar import Repository
    from hangar.records.summarize import list_history

    cmtList = []
    co = repo.checkout(write=True)
    co.add_ndarray_column(name='writtenaset', shape=(5, 7), dtype=np.float32)
    for cIdx in range(nCommits):
        if cIdx != 0:
            co = repo.checkout(write=True)
        sampList = []
        with co.columns['writtenaset'] as d:
            for prevKey in list(d.keys())[1:]:
                del d[prevKey]
            for sIdx in range(nSamples):
                arr = np.random.randn(*array5by7.shape).astype(np.float32) * 100
                d[str(sIdx)] = arr
                sampList.append(arr)
        cmt = co.commit(f'commit number: {cIdx}')
        cmtList.append((cmt, sampList))
        co.close()
    masterHist = list_history(repo._env.refenv, repo._env.branchenv, branch_name='master')

    repo.remote.add('origin', server_instance)
    push1 = repo.remote.push('origin', 'master')
    assert push1 == 'master'

    new_tmpdir = pjoin(managed_tmpdir, 'new')
    mkdir(new_tmpdir)
    newRepo = Repository(path=new_tmpdir, exists=False)
    newRepo.clone('Test User', 'tester@foo.com', server_instance, remove_old=True)
    assert newRepo.list_branches() == ['master', 'origin/master']
    for cmt, sampList in cmtList:
        with pytest.warns(UserWarning):
            nco = newRepo.checkout(commit=cmt)
        assert len(nco.columns) == 1
        assert 'writtenaset' in nco.columns
        assert len(nco.columns['writtenaset']) == len(sampList)

        assert nco.columns['writtenaset'].contains_remote_references is True
        remoteKeys = nco.columns['writtenaset'].remote_reference_keys
        assert tuple([str(idx) for idx in range(len(sampList))]) == remoteKeys
        for idx, _ in enumerate(sampList):
            sIdx = str(idx)
            assert sIdx in nco.columns['writtenaset']
            with pytest.raises(FileNotFoundError):
                shouldNotExist = nco.columns['writtenaset'][sIdx]
        nco.close()
    cloneMasterHist = list_history(newRepo._env.refenv, newRepo._env.branchenv, branch_name='master')
    assert cloneMasterHist == masterHist
    newRepo._env._close_environments()


@pytest.mark.parametrize('nMasterCommits,nMasterSamples', [[1, 4], [5, 10]])
@pytest.mark.parametrize('nDevCommits,nDevSamples', [[1, 3], [3, 5]])
def test_server_push_second_branch_with_new_commit(server_instance, repo,
                                                   array5by7, nMasterCommits,
                                                   nMasterSamples, nDevCommits,
                                                   nDevSamples):

    masterCmtList, devCmtList = [], []
    co = repo.checkout(write=True)
    co.add_ndarray_column(name='writtenaset', shape=(5, 7), dtype=np.float32)
    for cIdx in range(nMasterCommits):
        if cIdx != 0:
            co = repo.checkout(write=True)
        masterSampList = []
        with co.columns['writtenaset'] as d:
            for prevKey in list(d.keys())[1:]:
                del d[prevKey]
            for sIdx in range(nMasterSamples):
                arr = np.random.randn(*array5by7.shape).astype(np.float32) * 100
                d[str(sIdx)] = arr
                masterSampList.append(arr)
        cmt = co.commit(f'master commit number: {cIdx}')
        masterCmtList.append((cmt, masterSampList))
        co.close()

    repo.remote.add('origin', server_instance)
    push1 = repo.remote.push('origin', 'master')
    assert push1 == 'master'

    branch = repo.create_branch('testbranch')
    for cIdx in range(nDevCommits):
        co = repo.checkout(write=True, branch=branch.name)
        devSampList = []
        with co.columns['writtenaset'] as d:
            for prevKey in list(d.keys())[1:]:
                del d[prevKey]
            for sIdx in range(nDevSamples):
                arr = np.random.randn(*array5by7.shape).astype(np.float32) * 100
                d[str(sIdx)] = arr
                devSampList.append(arr)
        cmt = co.commit(f'dev commit number: {cIdx}')
        devCmtList.append((cmt, devSampList))
        co.close()

    push2 = repo.remote.push('origin', branch.name)
    assert push2 == branch.name


@pytest.mark.parametrize('nMasterCommits,nMasterSamples', [[1, 4], [5, 10]])
@pytest.mark.parametrize('nDevCommits,nDevSamples', [[1, 5], [3, 5]])
def test_server_push_second_branch_with_new_commit_then_clone_partial_fetch(
        server_instance, repo, managed_tmpdir, array5by7, nMasterCommits,
        nMasterSamples, nDevCommits, nDevSamples):
    from hangar import Repository
    from hangar.records.summarize import list_history

    # Push master branch test
    masterCmtList = []
    co = repo.checkout(write=True)
    co.add_ndarray_column(name='writtenaset', shape=(5, 7), dtype=np.float32)
    for cIdx in range(nMasterCommits):
        if cIdx != 0:
            co = repo.checkout(write=True)
        masterSampList = []
        with co.columns['writtenaset'] as d:
            for prevKey in list(d.keys())[1:]:
                del d[prevKey]
            for sIdx in range(nMasterSamples):
                arr = np.random.randn(*array5by7.shape).astype(np.float32) * 100
                d[str(sIdx)] = arr
                masterSampList.append(arr)
        cmt = co.commit(f'master commit number: {cIdx}')
        masterCmtList.append((cmt, masterSampList))
        co.close()

    repo.remote.add('origin', server_instance)
    push1 = repo.remote.push('origin', 'master')
    assert push1 == 'master'
    masterHist = list_history(repo._env.refenv, repo._env.branchenv, branch_name='master')

    # Push dev branch test
    devCmtList = []
    branch = repo.create_branch('testbranch')
    for cIdx in range(nDevCommits):
        co = repo.checkout(write=True, branch=branch.name)
        devSampList = []
        with co.columns['writtenaset'] as d:
            for prevKey in list(d.keys())[1:]:
                del d[prevKey]
            for sIdx in range(nDevSamples):
                arr = np.random.randn(*array5by7.shape).astype(np.float32) * 100
                d[str(sIdx)] = arr
                devSampList.append(arr)
        cmt = co.commit(f'dev commit number: {cIdx}')
        devCmtList.append((cmt, devSampList))
        co.close()

    push2 = repo.remote.push('origin', branch.name)
    assert push2 == branch.name
    branchHist = list_history(repo._env.refenv, repo._env.branchenv, branch_name=branch.name)

    # Clone test (master branch)
    new_tmpdir = pjoin(managed_tmpdir, 'new')
    mkdir(new_tmpdir)
    newRepo = Repository(path=new_tmpdir, exists=False)
    newRepo.clone('Test User', 'tester@foo.com', server_instance, remove_old=True)
    assert newRepo.list_branches() == ['master', 'origin/master']
    for cmt, sampList in masterCmtList:
        with pytest.warns(UserWarning):
            nco = newRepo.checkout(commit=cmt)
        assert len(nco.columns) == 1
        assert 'writtenaset' in nco.columns
        assert len(nco.columns['writtenaset']) == nMasterSamples

        assert nco.columns['writtenaset'].contains_remote_references is True
        remoteKeys = nco.columns['writtenaset'].remote_reference_keys
        assert tuple([str(idx) for idx in range(len(sampList))]) == remoteKeys
        for idx, _ in enumerate(sampList):
            sIdx = str(idx)
            assert sIdx in nco.columns['writtenaset']
            with pytest.raises(FileNotFoundError):
                shouldNotExist = nco.columns['writtenaset'][sIdx]
        nco.close()
    cloneMasterHist = list_history(newRepo._env.refenv, newRepo._env.branchenv, branch_name='master')
    assert cloneMasterHist == masterHist

    # Fetch test
    fetch = newRepo.remote.fetch('origin', branch=branch.name)
    assert fetch == f'origin/{branch.name}'
    assert newRepo.list_branches() == ['master', 'origin/master', f'origin/{branch.name}']
    for cmt, sampList in devCmtList:

        with pytest.warns(UserWarning):
            nco = newRepo.checkout(commit=cmt)
        assert len(nco.columns) == 1
        assert 'writtenaset' in nco.columns
        assert len(nco.columns['writtenaset']) == nDevSamples

        assert nco.columns['writtenaset'].contains_remote_references is True
        remoteKeys = nco.columns['writtenaset'].remote_reference_keys
        assert tuple([str(idx) for idx in range(len(sampList))]) == remoteKeys
        for idx, _ in enumerate(sampList):
            sIdx = str(idx)
            assert sIdx in nco.columns['writtenaset']
            with pytest.raises(FileNotFoundError):
                shouldNotExist = nco.columns['writtenaset'][sIdx]
        nco.close()

    cloneBranchHist = list_history(newRepo._env.refenv, newRepo._env.branchenv, branch_name=f'origin/{branch.name}')
    assert cloneBranchHist == branchHist
    newRepo._env._close_environments()


@pytest.mark.filterwarnings('ignore::UserWarning')
@pytest.mark.parametrize('nMasterCommits,nMasterSamples', [[2, 10]])
@pytest.mark.parametrize('nDevCommits,nDevSamples', [[1, 16]])
@pytest.mark.parametrize('fetchBranch,fetchCommit,fetchAsetns,fetchNbytes,fetchAll_history', [
    ['master',      None,  None,              None,  False],
    ['testbranch',  None,  None,              None,  False],
    [None,          'ma',  None,              None,  False],
    [None,          'br',  None,              None,  False],
    ['master',      None,  ('writtenaset',),  None,  False],
    ['testbranch',  None,  ('_two',),         None,  False],
    [None,          'ma',  ('writtenaset',),  None,  False],
    [None,          'br',  ('_two',),         None,  False],
    ['master',      None,  None,              None,  False],
    ['testbranch',  None,  None,              None,  False],
    [None,          'ma',  None,              None,  False],
    [None,          'br',  None,              None,  False],
    ['master',      None,  ('writtenaset',),  None,  False],
    ['testbranch',  None,  ('_two',),         None,  False],
    [None,          'ma',  ('writtenaset',),  None,  False],
    [None,          'br',  ('_two',),         None,  False],
    ['master',      None,  None,              None,  True],
    ['testbranch',  None,  None,              None,  True],
    [None,          'ma',  None,              None,  True],
    [None,          'br',  None,              None,  True],
    ['master',      None,  ('writtenaset',),  None,  True],
    ['testbranch',  None,  ('_two',),         None,  True],
    [None,          'ma',  ('writtenaset',),  None,  True],
    [None,          'br',  ('_two',),         None,  True],
    ['master',      None,  None,              1000,  False],
    ['testbranch',  None,  None,              1000,  False],
    [None,          'ma',  None,              1000,  False],
    [None,          'br',  None,              1000,  False],
    ['master',      None,  ('writtenaset',),  1000,  False],
    ['testbranch',  None,  ('_two',),         1000,  False],
    [None,          'ma',  ('writtenaset',),  1000,  False],
    [None,          'br',  ('_two',),         1000,  False],
    [None,          'br',  ('_two',),         1000,  True],  # will raise error
])
def test_server_push_two_branch_then_clone_fetch_data_options(
        server_instance, repo, managed_tmpdir, array5by7, nMasterCommits,
        nMasterSamples, nDevCommits, nDevSamples, fetchBranch, fetchCommit,
        fetchAsetns, fetchNbytes, fetchAll_history):
    from hangar import Repository
    from hangar.records.summarize import list_history

    # Push master branch test
    masterCmts = {}
    co = repo.checkout(write=True)
    co.add_ndarray_column(name='writtenaset', shape=(5, 7), dtype=np.float32)
    co.add_ndarray_column(name='_two', shape=(20), dtype=np.float32)
    for cIdx in range(nMasterCommits):
        if cIdx != 0:
            co = repo.checkout(write=True)
        masterSampList1 = []
        masterSampList2 = []
        with co.columns['writtenaset'] as d, co.columns['_two'] as dd:
            for prevKey in list(d.keys())[1:]:
                del d[prevKey]
                del dd[prevKey]

            for sIdx in range(nMasterSamples):
                arr1 = np.random.randn(*array5by7.shape).astype(np.float32) * 100
                d[str(sIdx)] = arr1
                masterSampList1.append(arr1)
                arr2 = np.random.randn(20).astype(np.float32)
                dd[str(sIdx)] = arr2
                masterSampList2.append(arr2)
        cmt = co.commit(f'master commit number: {cIdx}')
        masterCmts[cmt] = (masterSampList1, masterSampList2)
        co.close()

    repo.remote.add('origin', server_instance)
    push1 = repo.remote.push('origin', 'master')
    assert push1 == 'master'
    masterHist = list_history(repo._env.refenv, repo._env.branchenv, branch_name='master')

    # Push dev branch test
    devCmts = masterCmts.copy()
    branch = repo.create_branch('testbranch')
    for cIdx in range(nDevCommits):
        co = repo.checkout(write=True, branch=branch.name)
        devSampList1 = []
        devSampList2 = []
        with co.columns['writtenaset'] as d, co.columns['_two'] as dd:
            for prevKey in list(d.keys())[1:]:
                del d[prevKey]
                del dd[prevKey]

            for sIdx in range(nDevSamples):
                arr1 = np.random.randn(*array5by7.shape).astype(np.float32) * 100
                d[str(sIdx)] = arr1
                devSampList1.append(arr1)
                arr2 = np.random.randn(20).astype(np.float32)
                dd[str(sIdx)] = arr2
                devSampList2.append(arr2)
        cmt = co.commit(f'dev commit number: {cIdx}')
        devCmts[cmt] = (devSampList1, devSampList2)
        co.close()

    push2 = repo.remote.push('origin', branch.name)
    assert push2 == branch.name
    branchHist = list_history(repo._env.refenv, repo._env.branchenv, branch_name=branch.name)

    # -------------------------- end setup ------------------------------------

    # Clone test (master branch)
    new_tmpdir = pjoin(managed_tmpdir, 'new')
    mkdir(new_tmpdir)
    newRepo = Repository(path=new_tmpdir, exists=False)
    newRepo.clone('Test User', 'tester@foo.com', server_instance, remove_old=True)
    newRepo.remote.fetch('origin', branch=branch.name)
    newRepo.create_branch('testbranch', base_commit=branchHist['head'])
    assert newRepo.list_branches() == ['master', 'origin/master', f'origin/{branch.name}', branch.name]

    # ------------------ format arguments dependingon options -----------------

    kwargs = {
        'column_names': fetchAsetns,
        'max_num_bytes': fetchNbytes,
        'retrieve_all_history': fetchAll_history,
    }
    if fetchBranch is not None:
        func = branchHist if fetchBranch == 'testbranch' else masterHist
        kwargs['branch'] = fetchBranch
        kwargs['commit'] = None
    else:
        func = branchHist if fetchBranch == 'br' else masterHist
        kwargs['branch'] = None
        kwargs['commit'] = func['head']

    if fetchAll_history is True:
        commits_to_check = func['order']
    else:
        commits_to_check = [func['head']]

    # ----------------------- retrieve data with desired options --------------

    # This case should fail
    if (fetchAll_history is True) and isinstance(fetchNbytes, int):
        try:
            with pytest.raises(ValueError):
                fetch_commits = newRepo.remote.fetch_data(remote='origin', **kwargs)
        finally:
            newRepo._env._close_environments()
        return True
    # get data
    fetch_commits = newRepo.remote.fetch_data(remote='origin', **kwargs)
    assert commits_to_check == fetch_commits

    # ------------- check that you got everything you expected ----------------

    for fCmt in fetch_commits:
        co = newRepo.checkout(commit=fCmt)
        assert co.commit_hash == fCmt

        # when we are checking one aset only
        if isinstance(fetchAsetns, tuple):
            d = co.columns[fetchAsetns[0]]
            # ensure we didn't fetch the other data simultaneously

            ds1SampList, ds2SampList = devCmts[fCmt]
            if fetchAsetns[0] == 'writtenaset':
                compare = ds1SampList
            else:
                compare = ds2SampList

            totalSeen = 0
            for idx, samp in enumerate(compare):
                if fetchNbytes is None:
                    assert np.allclose(samp, d[str(idx)])
                else:
                    try:
                        arr = d[str(idx)]
                        assert np.allclose(samp, arr)
                        totalSeen += arr.nbytes
                    except FileNotFoundError:
                        pass
                    assert totalSeen <= fetchNbytes

        # compare both asets at the same time
        else:
            d = co.columns['writtenaset']
            dd = co.columns['_two']
            ds1List, ds2List = devCmts[fCmt]
            totalSeen = 0
            for idx, ds1ds2 in enumerate(zip(ds1List, ds2List)):
                ds1, ds2 = ds1ds2
                if fetchNbytes is None:
                    assert np.allclose(ds1, d[str(idx)])
                    assert np.allclose(ds2, dd[str(idx)])
                else:
                    try:
                        arr1 = d[str(idx)]
                        assert np.allclose(ds1, arr1)
                        totalSeen += arr1.nbytes
                    except FileNotFoundError:
                        pass
                    try:
                        arr2 = dd[str(idx)]
                        assert np.allclose(ds2, arr2)
                        totalSeen += arr2.nbytes
                    except FileNotFoundError:
                        pass
                    assert totalSeen <= fetchNbytes
        co.close()
    newRepo._env._close_environments()


def test_push_unchanged_repo_makes_no_modifications(written_two_cmt_server_repo):
    _, repo = written_two_cmt_server_repo
    with pytest.warns(UserWarning):
        branchName = repo.remote.push('origin', 'master')
    assert branchName == 'master'


def test_fetch_unchanged_repo_makes_no_modifications(written_two_cmt_server_repo):
    _, repo = written_two_cmt_server_repo
    with pytest.warns(UserWarning):
        branchName = repo.remote.fetch('origin', 'master')
    assert branchName == 'master'


def test_fetch_newer_disk_repo_makes_no_modifications(written_two_cmt_server_repo):
    _, repo = written_two_cmt_server_repo
    co = repo.checkout(write=True)
    co.add_str_column('test_meta')
    co['test_meta'][0] = 'lol'
    co.commit('newer commit')
    co.close()
    with pytest.warns(UserWarning):
        branchName = repo.remote.fetch('origin', 'master')
    assert branchName == 'master'


def test_fetch_branch_which_does_not_exist_client_server_raises_rpc_error(written_two_cmt_server_repo):
    import grpc
    _, repo = written_two_cmt_server_repo
    with pytest.raises(grpc.RpcError) as rpc_error:
        repo.remote.fetch('origin', 'not-a-branch')
    assert rpc_error.value._state.code == grpc.StatusCode.NOT_FOUND


def test_fetch_branch_on_client_which_does_not_existserver_raises_rpc_error(written_two_cmt_server_repo):
    import grpc
    _, repo = written_two_cmt_server_repo
    repo.create_branch('new-branch')
    with pytest.raises(grpc.RpcError) as exc_info:
        repo.remote.fetch('origin', 'new-branch')
    assert exc_info.value._state.code == grpc.StatusCode.NOT_FOUND


def test_push_clone_three_way_merge(server_instance, repo_2_br_no_conf, managed_tmpdir):
    from hangar import Repository

    repo_2_br_no_conf.remote.add('origin', server_instance)
    push1 = repo_2_br_no_conf.remote.push('origin', 'master')
    assert push1 == 'master'
    push2 = repo_2_br_no_conf.remote.push('origin', 'testbranch')
    assert push2 == 'testbranch'

    test_head = repo_2_br_no_conf.log(branch='testbranch', return_contents=True)['head']
    master_head = repo_2_br_no_conf.log(branch='master', return_contents=True)['head']

    merge_cmt = repo_2_br_no_conf.merge('merge commit', 'master', 'testbranch')
    merge_head = repo_2_br_no_conf.log(branch='master', return_contents=True)['head']
    merge_order = repo_2_br_no_conf.log(branch='master', return_contents=True)['order']
    merge_push = repo_2_br_no_conf.remote.push('origin', 'master')
    assert merge_push == 'master'
    assert merge_head != master_head
    assert merge_head != test_head

    new_tmpdir = pjoin(managed_tmpdir, 'new')
    mkdir(new_tmpdir)
    newRepo = Repository(path=new_tmpdir, exists=False)
    newRepo.clone('Test User', 'tester@foo.com', server_instance, remove_old=True)

    clone_head = newRepo.log(branch='master', return_contents=True)['head']
    clone_order = newRepo.log(branch='master', return_contents=True)['order']
    assert clone_head == merge_head == merge_cmt
    assert merge_order == clone_order
    newRepo._env._close_environments()


# ---------------------------- fixture func servers ---------------------------


@pytest.fixture()
def server_instance_nbytes_limit(monkeypatch, managed_tmpdir, worker_id, mocker):
    from hangar.remote.server import serve

    address = f'localhost:{randint(50000, 59999)}'
    base_tmpdir = pjoin(managed_tmpdir, f'{worker_id[-1]}')
    mkdir(base_tmpdir)
    server, hangserver, _ = serve(base_tmpdir, overwrite=True, channel_address=address)
    hangserver.CFG['SERVER_GRPC']['fetch_max_nbytes'] = '500000'
    hangserver.CFG['CLIENT_GRPC']['push_max_nbytes'] = '500000'
    server.start()
    yield address

    hangserver.env._close_environments()
    server.stop(0.05)
    time.sleep(0.1)
    if platform.system() == 'Windows':
        # time for open file handles to close before tmp dir can be removed.
        time.sleep(0.1)


@pytest.fixture()
def server_instance_push_restricted(managed_tmpdir, worker_id):
    from hangar.remote.server import serve

    address = f'localhost:{randint(50000, 59999)}'
    base_tmpdir = pjoin(managed_tmpdir, f'{worker_id[-1]}')
    mkdir(base_tmpdir)
    server, hangserver, _ = serve(base_tmpdir,
                                  overwrite=True,
                                  channel_address=address,
                                  restrict_push=True,
                                  username='right_username',
                                  password='right_password')
    server.start()
    yield address

    hangserver.env._close_environments()
    server.stop(0.05)
    time.sleep(0.1)
    if platform.system() == 'Windows':
        # time for open file handles to close before tmp dir can be removed.
        time.sleep(0.1)


# -----------------------------------------------------------------------------

@pytest.mark.skip(reason='unknown test failures intermitently')
def test_push_clone_digests_exceeding_server_nbyte_limit(mocker, server_instance_nbytes_limit, repo, managed_tmpdir):
    from hangar import Repository
    from hangar.remote import chunks, client

    # Push master branch test
    masterCmtList = []
    co = repo.checkout(write=True)
    co.add_ndarray_column(name='aset', shape=(50, 50), dtype=np.float32)
    for cIdx in range(4):
        if cIdx != 0:
            co = repo.checkout(write=True)
        masterSampList = []
        with co.columns['aset'] as d:
            for prevKey in list(d.keys())[1:]:
                del d[prevKey]
            for sIdx in range(70):
                arr = np.random.randint(0, 255, size=(50, 50)).astype(np.float32)
                d[str(sIdx)] = arr
                masterSampList.append(arr)
        cmt = co.commit(f'master commit number: {cIdx}')
        masterCmtList.append((cmt, masterSampList))
        co.close()

    repo.remote.add('origin', server_instance_nbytes_limit)

    spy = mocker.spy(chunks, 'tensorChunkedIterator')
    push1 = repo.remote.push('origin', 'master')
    assert chunks.tensorChunkedIterator.call_count == 6
    for call in spy.call_args_list:
        assert call[1]['uncomp_nbytes'] <= 550_000 # maximum amount over 100_000 observed in test development

    assert push1 == 'master'

    # Clone test (master branch)
    new_tmpdir = pjoin(managed_tmpdir, 'new')
    mkdir(new_tmpdir)
    newRepo = Repository(path=new_tmpdir, exists=False)
    newRepo.clone('Test User', 'tester@foo.com', server_instance_nbytes_limit, remove_old=True)
    assert newRepo.list_branches() == ['master', 'origin/master']

    spy = mocker.spy(client.HangarClient, 'fetch_data')
    for cmt, sampList in masterCmtList:
        newRepo.remote.fetch_data('origin', commit=cmt)
        nco = newRepo.checkout(commit=cmt)
        assert len(nco.columns) == 1
        assert 'aset' in nco.columns
        assert len(nco.columns['aset']) == 70
        for sIdx, samp in enumerate(sampList):
            assert np.allclose(nco.columns['aset'][str(sIdx)], samp)
        nco.close()
        del nco
    assert client.HangarClient.fetch_data.call_count == 8
    newRepo._env._close_environments()


def test_push_restricted_with_right_username_password(server_instance_push_restricted, repo, managed_tmpdir):
    from hangar import Repository

    # Push master branch test
    masterCmtList = []
    co = repo.checkout(write=True)
    co.add_ndarray_column(name='aset', shape=(50, 20), dtype=np.float32)
    for cIdx in range(1):
        if cIdx != 0:
            co = repo.checkout(write=True)
        masterSampList = []
        with co.columns['aset'] as d:
            for prevKey in list(d.keys())[1:]:
                del d[prevKey]
            for sIdx in range(70):
                arr = np.random.randn(50, 20).astype(np.float32)
                d[str(sIdx)] = arr
                masterSampList.append(arr)
        cmt = co.commit(f'master commit number: {cIdx}')
        masterCmtList.append((cmt, masterSampList))
        co.close()

    repo.remote.add('origin', server_instance_push_restricted)
    push1 = repo.remote.push('origin',
                             'master',
                             username='right_username',
                             password='right_password')
    assert push1 == 'master'

    # Clone test (master branch)
    new_tmpdir = pjoin(managed_tmpdir, 'new')
    mkdir(new_tmpdir)
    newRepo = Repository(path=new_tmpdir, exists=False)
    newRepo.clone('Test User', 'tester@foo.com', server_instance_push_restricted, remove_old=True)
    assert newRepo.list_branches() == ['master', 'origin/master']
    for cmt, sampList in masterCmtList:
        newRepo.remote.fetch_data('origin', commit=cmt)
        nco = newRepo.checkout(commit=cmt)
        assert len(nco.columns) == 1
        assert 'aset' in nco.columns
        assert len(nco.columns['aset']) == 70
        for sIdx, samp in enumerate(sampList):
            assert np.allclose(nco.columns['aset'][str(sIdx)], samp)
        nco.close()
    newRepo._env._close_environments()


def test_push_restricted_wrong_user_and_password(server_instance_push_restricted, repo, managed_tmpdir):

    # Push master branch test
    masterCmtList = []
    co = repo.checkout(write=True)
    co.add_ndarray_column(name='aset', shape=(50, 20), dtype=np.float32)
    for cIdx in range(1):
        if cIdx != 0:
            co = repo.checkout(write=True)
        masterSampList = []
        with co.columns['aset'] as d:
            for prevKey in list(d.keys())[1:]:
                del d[prevKey]
            for sIdx in range(70):
                arr = np.random.randn(50, 20).astype(np.float32)
                d[str(sIdx)] = arr
                masterSampList.append(arr)
        cmt = co.commit(f'master commit number: {cIdx}')
        masterCmtList.append((cmt, masterSampList))
        co.close()

    repo.remote.add('origin', server_instance_push_restricted)
    with pytest.raises(PermissionError):
        push1 = repo.remote.push('origin',
                                 'master',
                                 username='wrong_username',
                                 password='right_password')

    with pytest.raises(PermissionError):
        push2 = repo.remote.push('origin',
                                 'master',
                                 username='right_username',
                                 password='wrong_password')

    with pytest.raises(PermissionError):
        push3 = repo.remote.push('origin',
                                 'master',
                                 username='wrong_username',
                                 password='wrong_password')



"""
High level methods let user interact with hangar without diving into the internal
methods of hangar. We have enabled four basic entry points as high level methods

1. :func:`.load`
2. :func:`.save`
3. :func:`.show`
4. :func:`.board_show`

These entry points by itself is not capable of doing anything. But they are entry
points to the same methods in the `hangar.external` plugins available in pypi. These
high level entry points are used by the CLI for doing import, export and view
operations as well as the `hangarboard <https://github.com/tensorwerk/hangarboard>`_
for visualization (using ``board_show``)
"""
from typing import Tuple

import numpy as np

from .plugin_manager import PluginManager


pm = PluginManager()


def load(fpath: str,
         plugin: str = None,
         extension: str = None,
         **plugin_kwargs) -> Tuple[np.ndarray, str]:
    """
    Wrapper to load data from file into memory as numpy arrays using
    plugin's `load` method

    Parameters
    ----------
    fpath : str
        Data file path, e.g. ``path/to/test.jpg``
    plugin : str, optional
        Name of plugin to use.  By default, the preferred plugin for the
        given file format tried until a suitable. This cannot be `None` if
        `extension` is also `None`
    extension : str, optional
        Format of the file. This is used to infer which plugin to use
        in case plugin name is not provided. This cannot be `None` if
        `plugin` is also `None`

    Other Parameters
    ----------------
    plugin_kwargs : dict
        Plugin specific keyword arguments. If the function is being called from
        command line argument, all the unknown keyword arguments will be collected
        as ``plugin_kwargs``

    Returns
    -------
    img_array : :class:`numpy.ndarray`
        data returned from the given plugin.

    """
    if not pm.plugins_loaded:
        pm.reset_plugins()
    func = pm.get_plugin('load', plugin=plugin, extension=extension)
    return func(fpath, **plugin_kwargs)


def save(arr: np.ndarray, outdir: str, sample_det: str, extension: str,
         plugin: str = None, **plugin_kwargs):
    """Wrapper plugin ``save`` methods which dump :class:`numpy.ndarray` to disk.

    Parameters
    ----------
    arr : :class:`numpy.ndarray`
        Numpy array to be saved to file
    outdir : str
        Target directory
    sample_det : str
        Sample name and type of the sample name formatted as
        ``sample_name_type:sample_name``
    extension : str
        Format of the file. This is used to infer which plugin to use in case
        plugin name is not provided. This cannot be ``None`` if ``plugin`` is
        also ``None``
    plugin : str, optional
        Name of plugin to use.  By default, the preferred plugin for the given
        file format tried until a suitable. This cannot be ``None`` if
        ``extension`` is also ``None``

    Other Parameters
    ----------------
    plugin_kwargs : dict
        Plugin specific keyword arguments. If the function is being called from
        command line argument, all the unknown keyword arguments will be
        collected as ``plugin_kwargs``

    Notes
    -----
    CLI or this method does not create the file name where to save. Instead
    they pass the required details downstream to the plugins to do that once
    they verify the given ``outdir`` is a valid directory. It is because we
    expect to get data entries where one data entry is one file (like images)
    and also data entries where multiple entries goes to single file (like
    CSV). With these ambiguous cases in hand, it's more sensible to let the
    plugin handle the file handling accordingly.
    """
    if not pm.plugins_loaded:
        pm.reset_plugins()
    func = pm.get_plugin('save', plugin=plugin, extension=extension)
    func(arr, outdir, sample_det, extension, **plugin_kwargs)


def show(arr: np.ndarray, plugin: str = None,
         extension: str = None, **plugin_kwargs):  # pragma: no cover
    """Wrapper to display :class:`numpy.ndarray` via plugin ``show`` method.

    Parameters
    ----------
    arr : :class:`numpy.ndarray`
        Data to process into some human understandable representation.
    plugin : str, optional
        Name of plugin to use.  By default, the preferred plugin for the
        given file format tried until a suitable. This cannot be ``None`` if
        ``extension`` is also ``None``
    extension : str, optional
        Format of the file. This is used to infer which plugin to use
        in case plugin name is not provided. This cannot be ``None`` if
        ``plugin`` is also ``None``

    Other Parameters
    ----------------
    plugin_kwargs : dict
        Plugin specific keyword arguments. If the function is being called from
        command line argument, all the unknown keyword arguments will be
        collected as ``plugin_kwargs``
    """
    if not pm.plugins_loaded:
        pm.reset_plugins()
    func = pm.get_plugin('show', plugin=plugin, extension=extension)
    return func(arr, **plugin_kwargs)


def board_show(arr: np.ndarray, plugin: str = None,
               extension: str = None, **plugin_kwargs):
    """
    Wrapper to convert the numpy array using the ``board_show`` method
    of the plugin to make it displayable in the web UI

    Parameters
    ----------
    arr : :class:`numpy.ndarray`
        Data to process into some human understandable representation.
    plugin : str, optional
        Name of plugin to use.  By default, the preferred plugin for the
        given file format tried until a suitable. This cannot be ``None`` if
        ``extension`` is also ``None``
    extension : str, optional
        Format of the file. This is used to infer which plugin to use
        in case plugin name is not provided. This cannot be ``None`` if
        ``plugin`` is also ``None``

    Other Parameters
    ----------------
    plugin_kwargs : dict
        Plugin specific keyword arguments. If the function is being called from
        command line argument, all the unknown keyword arguments will be
        collected as ``plugin_kwargs``
    """
    if not pm.plugins_loaded:
        pm.reset_plugins()
    func = pm.get_plugin('board_show', plugin=plugin, extension=extension)
    return func(arr, **plugin_kwargs)



import pytest
import numpy as np


from hangar.typesystem import NdarrayFixedShape, NdarrayVariableShape


class TestInvalidValues:

    @pytest.mark.parametrize('shape,expected_exc', [
        [tuple(range(32)), ValueError],
        [(1.2, 2), TypeError],
        [[1, 2], TypeError],
        ['shouldntwork', TypeError],
    ])
    def test_shape_not_tuple_of_int_less_than_32_dims(self, shape, expected_exc):
        with pytest.raises(expected_exc):
            NdarrayFixedShape(shape=shape, dtype=np.uint8, column_layout='flat')
        with pytest.raises(expected_exc):
            NdarrayVariableShape(shape=shape, dtype=np.uint8, column_layout='flat')

    @pytest.mark.parametrize(
        'coltype', ['str', str, 'notvalid', None, 32, 3.5, {'foo': 'bar'}, ascii])
    def test_column_type_must_be_ndarray(self, coltype):
        with pytest.raises(ValueError):
            NdarrayFixedShape(shape=(1,), dtype=np.uint8, column_layout='flat', column_type=coltype)
        with pytest.raises(ValueError):
            NdarrayVariableShape(shape=(1,), dtype=np.uint8, column_layout='flat', column_type=coltype)

    @pytest.mark.parametrize(
        'collayout', ['f', 'n', 'notvalid', None, 32, 3.5, {'foo': 'bar'}, ascii])
    def test_column_layout_must_be_valid_value(self, collayout):
        with pytest.raises(ValueError):
            NdarrayFixedShape(shape=(1,), dtype=np.uint8, column_layout=collayout)
        with pytest.raises(ValueError):
            NdarrayVariableShape(shape=(1,), dtype=np.uint8, column_layout=collayout)

    @pytest.mark.parametrize(
        'backend', ['30', 24, {'10': '10'}, ('00',), ['50', ], ascii, 'None'])
    def test_fixed_shape_backend_code_valid_value(self, backend):
        with pytest.raises(ValueError):
            NdarrayFixedShape(shape=(1,), dtype=np.uint8, column_layout='flat', backend=backend)

    @pytest.mark.parametrize(
        'backend', ['30', '01', 24, {'10': '10'}, ('00',), ['50', ], ascii, 'None'])
    def test_variable_shape_backend_code_valid_value(self, backend):
        with pytest.raises(ValueError):
            NdarrayVariableShape(shape=(1,), dtype=np.uint8, column_layout='flat', backend=backend)

    @pytest.mark.parametrize(
        'opts', ['val', [], (), [('key', 'val')], 10, ({'key': 'val'},), ascii])
    def test_backend_options_must_be_dict_or_nonetype(self, opts):
        with pytest.raises(TypeError):
            NdarrayFixedShape(shape=(1,), dtype=np.uint8, column_layout='flat', backend='00', backend_options=opts)
        with pytest.raises(TypeError):
            NdarrayVariableShape(shape=(1,), dtype=np.uint8, column_layout='flat', backend='00', backend_options=opts)

    def test_backend_must_be_specified_if_backend_options_provided(self):
        with pytest.raises(ValueError):
            NdarrayFixedShape(shape=(1,), dtype=np.uint8, column_layout='flat', backend_options={})
        with pytest.raises(ValueError):
            NdarrayVariableShape(shape=(1,), dtype=np.uint8, column_layout='flat', backend_options={})

    @pytest.mark.parametrize(
        'schema_type', ['fixed_shape', True, 'str', np.uint8, 3, ascii])
    def test_variable_shape_must_have_variable_shape_schema_type(self, schema_type):
        with pytest.raises(ValueError):
            NdarrayVariableShape(shape=(1,), dtype=np.uint8, column_layout='flat', schema_type=schema_type)

    @pytest.mark.parametrize(
        'schema_type', ['variable_shape', True, 'str', np.uint8, 3, ascii])
    def test_fixed_shape_must_have_fixed_shape_schema_type(self, schema_type):
        with pytest.raises(ValueError):
            NdarrayFixedShape(shape=(1,), dtype=np.uint8, column_layout='flat', schema_type=schema_type)





from collections import deque


cdef class SizedDict:
    """Sized dictionary"""

    def __init__(self, int maxsize=1000):
        self._data = dict()
        self._maxsize = maxsize
        self._stack = deque()
        self._stack_size = 0

    @property
    def maxsize(self):
        return self._maxsize

    def __repr__(self):
        return repr(self._data)

    def __contains__(self, key):
        """Return True if d has a key key, else False."""
        cdef bint res
        res = key in self._data
        return res

    def __getitem__(self, key):
        """Return the item of d with key key. Raises a KeyError if key
        is not in the map.
        """
        return self._data[key]

    def get(self, key, default=None):
        """Return the value for key if key is in the dictionary, else default.

        If default is not given, it defaults to None, so that this method
        never raises a KeyError.
        """
        return self._data.get(key, default)

    def __len__(self):
        """Return the number of items in the dictionary d.
        """
        return self._stack_size

    def __iter__(self):
        """Return an iterator over the keys of the dictionary.

        This is a shortcut for iter(d.keys()).
        """
        return iter(self.keys())

    def __setitem__(self, key, value):
        """Set d[key] to value
        """
        if self._stack_size >= self._maxsize:
            k_pop = self._stack.popleft()
            del self._data[k_pop]
            self._stack_size = self._stack_size - 1
        self._stack.append(key)
        self._data[key] = value
        self._stack_size = self._stack_size + 1

    def __delitem__(self, key):
        """Remove d[key] from d. Raises a KeyError if key is not in the map.
        """
        del self._data[key]
        self._stack.remove(key)
        self._stack_size = self._stack_size - 1

    def keys(self):
        """Return a new view of the dictionary’s keys."""
        return self._data.keys()

    def values(self):
        """Return a new view of the dictionary’s values."""
        return self._data.values()

    def items(self):
        """Return a new view of the dictionary’s items (``(key, value)`` pairs).
        """
        return self._data.items()

    def clear(self):
        """Remove all items from the dictionary.
        """
        self._stack.clear()
        self._data.clear()
        self._stack_size = 0

    def pop(self, key, default=None):
        """If key is in the dictionary, remove it and return its value,
        else return default.

        If default is not given and key is not in the dictionary, a KeyError is raised.
        """
        cdef bint has_default

        has_default = not bool(default is None)
        if key in self._data:
            val = self._data.pop(key)
            self._stack.remove(key)
            self._stack_size = self._stack_size - 1
        elif has_default:
            val = default
        return val

    def popitem(self):
        """Remove and return a (key, value) pair from the dictionary.
        Pairs are returned in LIFO order.

        popitem() is useful to destructively iterate over a dictionary,
        as often used in set algorithms. If the dictionary is empty, calling
        popitem() raises a KeyError.
        """
        cdef object lifo_key, lifo_val
        lifo_key = self._stack.pop()
        lifo_val = self._data.pop(lifo_key)
        self._stack_size = self._stack_size - 1
        return lifo_key, lifo_val

    def update(self, other):
        """Update the dictionary with the key/value pairs from other, overwriting
        existing keys. Return None.

        update() accepts either another dictionary object or an iterable of
        key/value pairs (as tuples or other iterables of length two). If keyword
        arguments are specified, the dictionary is then updated with those
        key/value pairs: d.update(red=1, blue=2).
        """
        if not isinstance(other, dict):
            other = dict(other)
        for k, v in other.items():
            self[k] = v

    def setdefault(self, key, default=None):
        """If key is in the dictionary, return its value. If not, insert key
        with a value of default and return default. default defaults to None.
        """
        try:
            return self._data[key]
        except KeyError:
            self[key] = default
            return default



import pytest


@pytest.mark.parametrize('arg,key,expected', [
    ['AAABBBCCC', None, ['A', 'B', 'C']],
    ['AAABbBCcC', str.lower, ['A', 'B', 'C']],
    ['ABACBACDA', None, ['A', 'B', 'C', 'D']],
    ['ABacBaCAd', str.upper, ['A', 'B', 'c', 'd']],
])
def test_unique_everseen(arg, key, expected):
    from hangar.utils import unique_everseen

    res = list(unique_everseen(arg, key=key))
    assert res == expected


@pytest.mark.parametrize('pth', [pytest.File, None, 123])
def test_valid_directory_path_errors_on_invalid_path_arg(pth):
    from hangar.utils import is_valid_directory_path
    with pytest.raises(TypeError, match='Path arg `p`'):
        is_valid_directory_path(pth)


def test_valid_directory_path_recognizes_not_a_directory(managed_tmpdir):
    from hangar.utils import is_valid_directory_path
    from pathlib import Path

    test_pth = Path(managed_tmpdir, 'test.txt').resolve()
    with test_pth.open('w+') as f:
        f.write('hello')
    with pytest.raises(NotADirectoryError):
        is_valid_directory_path(test_pth)


@pytest.mark.parametrize('arg,expected', [
    [1, '1.00 B'],
    [1234, '1.23 kB'],
    [12345678, '12.35 MB'],
    [1234567890, '1.23 GB'],
    [1234567890000, '1.23 TB'],
    [1234567890000000, '1.23 PB']
])
def test_format_bytes(arg, expected):
    from hangar.utils import format_bytes

    res = format_bytes(arg)
    assert res == expected


@pytest.mark.parametrize('arg,expected', [
    ['100', 100],
    ['100 MB', 100000000],
    ['100M', 100000000],
    ['5kB', 5000],
    ['5.4 kB', 5400],
    ['1kiB', 1024],
    ['1e6', 1000000],
    ['1e6 kB', 1000000000],
    ['MB', 1000000]
])
def test_parse_bytes(arg, expected):
    from hangar.utils import parse_bytes

    res = parse_bytes(arg)
    assert res == expected


@pytest.mark.parametrize('arg,expected', [
    [0, 2],
    [1, 2],
    [2, 2],
    [3, 3],
    [4, 5],
    [7, 7],
    [174, 179],
    [10065, 10067],
    [104721, 104723],
])
def test_find_next_prime(arg, expected):
    from hangar.utils import find_next_prime

    res = find_next_prime(arg)
    assert res == expected



import pytest


@pytest.mark.parametrize('name', [
    'dummy branch', 'origin/master', '\nmaster', '\\master', 'master\n'
    'master\r\n', 'master ', 1412, 'foo !', 'foo@', 'foo#', 'foo$', '(foo)',
    'VeryLongNameIsInvalidOver64CharactersNotAllowedVeryLongNameIsInva'])
def test_create_branch_fails_invalid_name(aset_samples_initialized_repo, name):
    repo = aset_samples_initialized_repo
    with pytest.raises(ValueError):
        repo.create_branch(name)


def test_list_branches_only_reports_master_upon_initialization(repo):
    branches = repo.list_branches()
    assert branches == ['master']


def test_cannot_create_new_branch_from_initialized_repo_with_no_commits(repo):
    with pytest.raises(RuntimeError):
        repo.create_branch('testbranch')


def test_can_create_new_branch_from_repo_with_one_commit(repo):
    co = repo.checkout(write=True)
    co.add_str_column('test_meta')
    expected_digest = co.commit('first')
    co.close()

    branchRes = repo.create_branch('testbranch')

    assert branchRes.name == 'testbranch'
    assert branchRes.digest == expected_digest


def test_cannot_duplicate_branch_name(aset_samples_initialized_repo):
    aset_samples_initialized_repo.create_branch('testbranch')
    with pytest.raises(ValueError):
        aset_samples_initialized_repo.create_branch('testbranch')


def test_create_multiple_branches_different_name_same_commit(aset_samples_initialized_repo):
    b1 = aset_samples_initialized_repo.create_branch('testbranch1')
    b2 = aset_samples_initialized_repo.create_branch('testbranch2')
    b3 = aset_samples_initialized_repo.create_branch('testbranch3')

    assert b1.digest == b2.digest
    assert b2.digest == b3.digest
    assert b3.digest == b1.digest
    assert aset_samples_initialized_repo.list_branches() == ['master', 'testbranch1', 'testbranch2', 'testbranch3']


def test_create_branch_by_specifying_base_commit(repo):

    co = repo.checkout(write=True)
    co.add_str_column('test_meta')
    co.commit('first commit')
    first_digest = co.commit_hash
    co['test_meta']['foo'] = 'bar'
    second_digest = co.commit('second')
    co['test_meta']['hello'] = 'world'
    third_digest = co.commit('third')
    co['test_meta']['zen'] = 'python'
    fourth_digest = co.commit('fourth')
    co.close()

    assert repo.list_branches() == ['master']

    secBranch = repo.create_branch('dev-second', base_commit=second_digest)
    assert secBranch.name == 'dev-second'
    assert secBranch.digest == second_digest

    co = repo.checkout(branch='dev-second')
    assert len(co['test_meta']) == 1
    assert co['test_meta']['foo'] == 'bar'
    co.close()


def test_remove_branch_works_when_commits_align(repo):
    co = repo.checkout(write=True)
    co.add_str_column('test_meta')
    co.commit('first')
    co['test_meta']['foo'] = 'bar'
    masterHEAD = co.commit('second')
    co.close()
    repo.create_branch('testdelete')

    assert repo.list_branches() == ['master', 'testdelete']

    removedBranch = repo.remove_branch('testdelete')
    assert removedBranch.name == 'testdelete'
    assert removedBranch.digest == masterHEAD
    assert repo.list_branches() == ['master']


def test_delete_branch_raises_runtime_error_when_history_not_merged(repo):
    co = repo.checkout(write=True)
    co.add_str_column('test_meta')
    co.commit('first')
    co['test_meta']['foo'] = 'bar'
    masterHEAD = co.commit('second')
    co.close()

    repo.create_branch('testdelete')
    co = repo.checkout(write=True, branch='testdelete')
    co['test_meta']['hello'] = 'world'
    thirdDigest = co.commit('third')
    co.close()

    # checkout master so staging area is not on branch
    co = repo.checkout(write=True, branch='master')
    co.close()

    assert repo.list_branches() == ['master', 'testdelete']
    with pytest.raises(RuntimeError):
        repo.remove_branch('testdelete')


def test_delete_branch_completes_when_history_not_merged_but_force_option_set(repo):
    co = repo.checkout(write=True)
    co.add_str_column('test_meta')
    co.commit('first')
    co['test_meta']['foo'] = 'bar'
    masterHEAD = co.commit('second')
    co.close()

    repo.create_branch('testdelete')
    co = repo.checkout(write=True, branch='testdelete')
    co['test_meta']['hello'] = 'world'
    thirdDigest = co.commit('third')
    co.close()

    # checkout master so staging area is not on branch
    co = repo.checkout(write=True, branch='master')
    co.close()
    assert repo.list_branches() == ['master', 'testdelete']

    removedBranch = repo.remove_branch('testdelete', force_delete=True)
    assert removedBranch.name == 'testdelete'
    assert removedBranch.digest == thirdDigest
    assert repo.list_branches() == ['master']


def test_delete_branch_raises_value_error_if_invalid_branch_name(repo):
    co = repo.checkout(write=True)
    co.add_str_column('test_meta')
    co.commit('first')
    co['test_meta']['foo'] = 'bar'
    masterHEAD = co.commit('second')
    co.close()

    repo.create_branch('testdelete')
    co = repo.checkout(write=True, branch='testdelete')
    co['test_meta']['hello'] = 'world'
    thirdDigest = co.commit('third')
    co.close()

    assert repo.list_branches() == ['master', 'testdelete']
    with pytest.raises(ValueError):
        repo.remove_branch('doesnotexist')
    with pytest.raises(ValueError):
        repo.remove_branch('origin/master')


def test_delete_branch_raises_permission_error_if_writer_lock_held(repo):
    co = repo.checkout(write=True)
    co.add_str_column('test_meta')
    co.commit('first')
    co['test_meta']['foo'] = 'bar'
    masterHEAD = co.commit('second')
    co.close()

    repo.create_branch('testdelete')
    co = repo.checkout(write=True, branch='testdelete')
    co['test_meta']['hello'] = 'world'
    thirdDigest = co.commit('third')
    co.close()

    # checkout master so staging area is not on branch
    co = repo.checkout(write=True, branch='master')
    with pytest.raises(PermissionError):
        repo.remove_branch('testdelete')
    assert repo.list_branches() == ['master', 'testdelete']
    co.close()


def test_delete_branch_raises_permission_error_if_branch_requested_is_staging_head(repo):
    co = repo.checkout(write=True)
    co.add_str_column('test_meta')
    co.commit('first')
    co['test_meta']['foo'] = 'bar'
    masterHEAD = co.commit('second')
    co.close()

    repo.create_branch('testdelete')
    co = repo.checkout(write=True, branch='testdelete')
    co['test_meta']['hello'] = 'world'
    thirdDigest = co.commit('third')
    co.close()

    with pytest.raises(PermissionError):
        repo.remove_branch('testdelete')
    assert repo.list_branches() == ['master', 'testdelete']


def test_delete_branch_raises_permission_error_if_only_one_branch_left(repo):
    co = repo.checkout(write=True)
    co.add_str_column('test_meta')
    co['test_meta']['foo'] = 'bar'
    masterHEAD = co.commit('second')
    co.close()

    assert repo.list_branches() == ['master']
    with pytest.raises(PermissionError):
        repo.remove_branch('master')
    assert repo.list_branches() == ['master']



import os
import pytest
from hangar import Repository


def test_imports():
    import hangar
    from hangar import Repository


def test_starting_up_repo_warns_should_exist_no_args(managed_tmpdir):
    with pytest.warns(UserWarning):
        repo = Repository(path=managed_tmpdir)
    repo.init(user_name='tester', user_email='foo@test.bar', remove_old=True)
    assert repo.list_branches() == ['master']
    assert os.path.isdir(repo._repo_path)
    assert str(repo._repo_path) == os.path.join(managed_tmpdir, '.hangar')
    co = repo.checkout(write=True)
    assert co.diff.status() == 'CLEAN'
    co.close()
    repo._env._close_environments()


def test_starting_up_repo_warns_should_exist_manual_args(managed_tmpdir):
    with pytest.warns(UserWarning):
        repo = Repository(path=managed_tmpdir, exists=True)
    repo.init(user_name='tester', user_email='foo@test.bar', remove_old=True)
    assert repo.list_branches() == ['master']
    assert os.path.isdir(repo._repo_path)
    assert str(repo._repo_path) == os.path.join(managed_tmpdir, '.hangar')
    co = repo.checkout(write=True)
    assert co.diff.status() == 'CLEAN'
    co.close()
    repo._env._close_environments()


def test_starting_up_repo_does_not_warn_not_exist_manual_args(managed_tmpdir):
    with pytest.warns(None) as warn_recs:
        repo = Repository(path=managed_tmpdir, exists=False)
    assert len(warn_recs) == 0

    repo.init(user_name='tester', user_email='foo@test.bar', remove_old=True)
    assert repo.list_branches() == ['master']
    assert os.path.isdir(repo._repo_path)
    assert str(repo._repo_path) == os.path.join(managed_tmpdir, '.hangar')
    co = repo.checkout(write=True)
    assert co.diff.status() == 'CLEAN'
    co.close()
    repo._env._close_environments()


def test_initial_read_checkout(managed_tmpdir):
    repo = Repository(path=managed_tmpdir, exists=False)
    repo.init(user_name='tester', user_email='foo@test.bar', remove_old=True)
    with pytest.raises(ValueError):
        repo.checkout()
    repo._env._close_environments()


def test_initial_arrayset(managed_tmpdir, randomsizedarray):
    repo = Repository(path=managed_tmpdir, exists=False)
    repo.init(user_name='tester', user_email='foo@test.bar', remove_old=True)

    w_checkout = repo.checkout(write=True)
    assert len(w_checkout.columns) == 0
    with pytest.raises(KeyError):
        w_checkout.columns['aset']
    aset = w_checkout.add_ndarray_column('aset', prototype=randomsizedarray)
    assert aset.column == 'aset'
    w_checkout.close()
    repo._env._close_environments()


def test_empty_commit(managed_tmpdir, caplog):
    repo = Repository(path=managed_tmpdir, exists=False)
    repo.init(user_name='tester', user_email='foo@test.bar', remove_old=True)
    w_checkout = repo.checkout(write=True)
    with pytest.raises(RuntimeError):
        w_checkout.commit('this is a merge message')
    w_checkout.close()
    repo._env._close_environments()


def test_cannot_operate_without_repo_init(managed_tmpdir):
    repo = Repository(path=managed_tmpdir, exists=False)

    with pytest.raises(RuntimeError):
        repo.writer_lock_held()
    with pytest.raises(RuntimeError):
        repo.checkout()
    with pytest.raises(RuntimeError):
        repo.writer_lock_held()
    with pytest.raises(RuntimeError):
        repo.log()
    with pytest.raises(RuntimeError):
        repo.summary()
    with pytest.raises(RuntimeError):
        repo.merge('fail', 'master', 'nonexistant')
    with pytest.raises(RuntimeError):
        repo.create_branch('test')
    with pytest.raises(RuntimeError):
        repo.list_branches()
    with pytest.raises(RuntimeError):
        repo.force_release_writer_lock()

    with pytest.raises(RuntimeError):
        repo.remote.add('origin', 'foo')
    with pytest.raises(RuntimeError):
        repo.remote.remove('origin')
    with pytest.raises(RuntimeError):
        repo.remote.fetch('origin', 'master')
    with pytest.raises(RuntimeError):
        repo.remote.fetch_data('origin', branch='master')
    with pytest.raises(RuntimeError):
        repo.remote.list_all()
    with pytest.raises(RuntimeError):
        repo.remote.ping('origin')
    with pytest.raises(RuntimeError):
        repo.remote.push('origin', 'master')
    with pytest.raises(RuntimeError):
        repo.remove_branch('master')

    with pytest.raises(RuntimeError):
        repo.path
    with pytest.raises(RuntimeError):
        repo.version
    with pytest.raises(RuntimeError):
        repo.writer_lock_held
    with pytest.raises(RuntimeError):
        repo.size_human
    with pytest.raises(RuntimeError):
        repo.size_nbytes

    assert repo._env.repo_is_initialized is False


def test_check_repo_size(repo_20_filled_samples):
    from hangar.utils import parse_bytes, folder_size

    expected_nbytes = folder_size(repo_20_filled_samples._repo_path, recurse=True)
    nbytes = repo_20_filled_samples.size_nbytes
    assert expected_nbytes == nbytes

    format_nbytes = repo_20_filled_samples.size_human
    # account for rounding when converting int to str.
    assert nbytes * 0.95 <= parse_bytes(format_nbytes) <= nbytes * 1.05


def test_force_release_writer_lock(managed_tmpdir, monkeypatch):

    repo = Repository(path=managed_tmpdir, exists=False)
    repo.init(user_name='tester', user_email='foo@test.bar', remove_old=True)
    co = repo.checkout(write=True)
    orig_lock = str(co._writer_lock)

    def mock_true(*args, **kwargs):
        return True

    # try to release the writer lock with a process which has different uid
    co._writer_lock = 'lololol'
    with pytest.raises(RuntimeError):
        monkeypatch.setattr(co, '_verify_alive', mock_true)
        monkeypatch.setattr(co._columns, '_destruct', mock_true)
        co.close()
    # replace, but rest of object is closed
    monkeypatch.setattr(co, '_writer_lock', orig_lock)
    monkeypatch.delattr(co._columns, '_destruct')
    co.close()
    repo._env._close_environments()


def test_force_release_writer_lock_works(managed_tmpdir):
    repo = Repository(path=managed_tmpdir, exists=False)
    repo.init(user_name='tester', user_email='foo@test.bar', remove_old=True)
    co = repo.checkout(write=True)

    # try to release the writer lock with a process which has different uid
    with pytest.warns(ResourceWarning):
        repo.force_release_writer_lock()

    co._writer_lock == 'LOCK_AVAILABLE'
    co.close()
    # replace, but rest of object is closed
    repo._env._close_environments()


def test_get_ecosystem_details(managed_tmpdir):
    repo = Repository(path=managed_tmpdir, exists=False)
    repo.init(user_name='tester', user_email='foo@test.bar', remove_old=True)
    eco = repo._ecosystem_details()
    assert isinstance(eco, dict)
    assert 'host' in eco
    assert 'packages' in eco
    for package_name, version in eco['packages']:
        assert version is not None
    repo._env._close_environments()


def test_inject_repo_version(monkeypatch):
    import hangar
    monkeypatch.setattr("hangar.__version__", '0.2.0')
    assert hangar.__version__ == '0.2.0'


def test_check_repository_version(aset_samples_initialized_repo):
    from hangar import __version__
    from pkg_resources import parse_version

    repo = aset_samples_initialized_repo
    assert repo.version == parse_version(__version__).public


def test_check_repository_software_version_startup(managed_tmpdir):
    from hangar import Repository, __version__
    from pkg_resources import parse_version

    repo = Repository(managed_tmpdir, exists=False)
    repo.init('test user', 'test@foo.bar', remove_old=True)
    repo._env._close_environments()

    nrepo = Repository(managed_tmpdir, exists=True)
    assert nrepo.initialized is True
    assert nrepo.version == parse_version(__version__).public
    nrepo._env._close_environments()


@pytest.mark.parametrize('repo_v,hangar_v', [
    ['0.2.0', '0.3.0'],
    ['0.2.0', '0.3.1rc1'],
    ['0.2.0', '0.3.1.dev0'],
    ['0.2.0', '0.3.1'],
    ['0.3.0', '0.4.1.dev0'],
    ['0.3.0', '0.4.1rc1'],
    ['0.3.0', '0.4.0'],
    ['0.3.0', '0.4.1'],
    ['0.4.0', '0.5.0.dev0'],
    ['0.4.0', '0.5.0rc1'],
    ['0.4.0', '0.5.0'],
    ['0.4.0', '0.5.1'],
    ['0.5.0.dev0', '0.4.0'],
    ['0.5.0.dev0', '0.4.1'],
    ['0.5.0', '0.4.1'],
])
def test_check_repository_software_version_fails_hangar_version(monkeypatch, managed_tmpdir, repo_v, hangar_v):
    import hangar
    monkeypatch.setattr("hangar.__version__", hangar_v)
    monkeypatch.setattr("hangar.context.__version__", hangar_v)
    from hangar import Repository
    from hangar.records.vcompat import set_repository_software_version

    repo = Repository(managed_tmpdir, exists=False)
    repo.init('test user', 'test@foo.bar', remove_old=True)
    # force writing of new software version. should trigger error on next read.
    set_repository_software_version(repo._env.branchenv, repo_v, overwrite=True)
    try:
        assert repo.version == repo_v
    finally:
        repo._env._close_environments()

    assert hangar.__version__ == hangar_v

    with pytest.raises(RuntimeError):
        Repository(managed_tmpdir, exists=True)


@pytest.mark.parametrize('futureVersion', ['1.0.0', '0.14.1', '0.15.0', '1.4.1'])
def test_check_repository_software_version_works_on_newer_hangar_version(managed_tmpdir, monkeypatch, futureVersion):
    from hangar import Repository

    repo = Repository(managed_tmpdir, exists=False)
    repo.init('test user', 'test@foo.bar', remove_old=True)
    old_version = repo.version
    # force writing of new software version. should trigger error on next read.
    repo._env._close_environments()

    import hangar
    monkeypatch.setattr(hangar, '__version__', futureVersion)
    nrepo = Repository(managed_tmpdir, exists=True)
    assert hangar.__version__ == futureVersion
    assert nrepo.version == old_version
    nrepo._env._close_environments()



import types
import sys

import wrapt


@wrapt.decorator
def writer_checkout_only(wrapped, instance, args, kwargs) -> types.MethodType:
    """Only allow a method to be called in a write-enable checkout.

    Parameters
    ----------
    wrapped
        bound method which is being called
    instance
        class object being operated on ie. ``instance is self``
        in both (equality and identify).
    args
        argument list passed to the method
    kwargs
        keyword args dict passed to the method.

    Returns
    -------
    types.MethodType
        If instance._mode == 'a' (write enabled checkout) then
        operation is allowed and pass through args and kwargs
        to the method as specified.

    Raises
    ------
    PermissionError
        If the checkout is opened in read-only mode, then deny
        ability to call and raise error explaining why to user.
    """
    try:
        if instance._mode == 'a':  # user facing classes hide attribute
            return wrapped(*args, **kwargs)
        else:
            err = (f'Method "{wrapped.__func__.__name__}" '
                   f'cannot be called in a read-only checkout.')
            raise PermissionError(err) from None

    except AttributeError:
        if instance.mode == 'a':  # internal classes don't hide attribute
            return wrapped(*args, **kwargs)
        else:
            err = (f'Method "{wrapped.__func__.__name__}" '
                   f'cannot be called in a read-only checkout.')
            raise PermissionError(err) from None


@wrapt.decorator
def reader_checkout_only(wrapped, instance, args, kwargs) -> types.MethodType:
    """Only allow a method to be called in a read-only checkout.

    Parameters
    ----------
    wrapped
        bound method which is being called
    instance
        class object being operated on ie. ``instance is self``
        in both (equality and identify).
    args
        argument list passed to the method
    kwargs
        keyword args dict passed to the method.

    Returns
    -------
    types.MethodType
        If instance._mode == 'r' (read-only checkout) then
        operation is allowed and pass through args and kwargs
        to the method as specified.

    Raises
    ------
    PermissionError
        If the checkout is opened in write-enabled mode, then deny
        ability to call and raise error explaining why to user.
    """
    try:
        if instance._mode == 'r':  # user facing classes hide attribute
            return wrapped(*args, **kwargs)
        else:
            err = (f'Method "{wrapped.__func__.__name__}" '
                   f'cannot be called in a write-enabled checkout.')
            raise PermissionError(err) from None

    except AttributeError:
        if instance.mode == 'r':  # internal classes don't hide attribute
            return wrapped(*args, **kwargs)
        else:
            err = (f'Method "{wrapped.__func__.__name__}" '
                   f'cannot be called in a write-enabled checkout.')
            raise PermissionError(err) from None


def tb_params_last_called(tb: types.TracebackType) -> dict:
    """Get parameters of the last function called before exception thrown.

    Parameters
    ----------
    tb : types.TracebackType
        traceback object returned as the third item from sys.exc_info()
        corresponding to an exception raised in the last stack frame.

    Returns
    -------
    dict
        parameters passed to the last function called before the exception was
        thrown.
    """
    while tb.tb_next:
        tb = tb.tb_next
    frame = tb.tb_frame
    code = frame.f_code
    argcount = code.co_argcount
    if code.co_flags & 4:  # *args
        argcount += 1
    if code.co_flags & 8:  # **kwargs
        argcount += 1
    names = code.co_varnames[:argcount]
    params = {}
    for name in names:
        params[name] = frame.f_locals.get(name, '<deleted>')
    return params


def report_corruption_risk_on_parsing_error(func):
    """Decorator adding try/except handling non-explicit exceptions.

    Explicitly raised RuntimeErrors generally point to corrupted data
    identified by a cryptographic hash mismatch. However, in order to get to
    the point where such quantities can be processes, a non-trivial amount of
    parsing machinery must be run. Should any error be thrown in the parse
    machinery due to corrupted values, this method raises the exception in a
    useful form; providing traceback context, likely root cause (displayed to
    users), and the offending arguments passed to the function which threw the
    error.
    """
    def wrapped(*args, **kwargs):
        try:
            func(*args, **kwargs)
        except RuntimeError as e:
            raise e
        except Exception as e:
            raise RuntimeError(
                f'Corruption detected during {func.__name__}. Most likely this is the '
                f'result of unparsable record values. Exception msg `{str(e)}`. Params '
                f'`{tb_params_last_called(sys.exc_info()[2])}`') from e
    return wrapped



"""Remote server location unknown backend, Identifier: ``REMOTE_50``

Backend Identifiers
===================

*  Backend: ``5``
*  Version: ``0``
*  Format Code: ``50``
*  Canonical Name: ``REMOTE_50``

Storage Method
==============

*  This backend merely acts to record that there is some data sample with some
   ``hash`` and ``schema_shape`` present in the repository. It does not store the
   actual data on the local disk, but indicates that if it should be retrieved,
   you need to ask the remote hangar server for it. Once present on the local
   disk, the backend locating info will be updated with one of the `local` data
   backend specifications.

Record Format
=============

Fields Recorded for Each Array
------------------------------

*  Format Code
*  Schema Hash

Separators used
---------------

* ``SEP_KEY: ":"``

Examples
--------

1)  Adding the first piece of data to a file:

    *  Schema Hash: "ae43A21a"

    ``Record Data => '50:ae43A21a'``

1)  Adding to a piece of data to a the middle of a file:

    *  Schema Hash: "ae43A21a"

    ``Record Data => '50:ae43A21a'``

Technical Notes
===============

*  The schema_hash field is required in order to allow effective placement of
   actual retrieved data into suitable sized collections on a ``fetch-data()``
   operation
"""
from pathlib import Path
from typing import Optional

from .specs import REMOTE_50_DataHashSpec
from ..op_state import writer_checkout_only, reader_checkout_only
from ..typesystem import EmptyDict, checkedmeta


# -------------------------------- Parser Implementation ----------------------

_FmtCode = '50'


def remote_50_encode(schema_hash: str = '') -> bytes:
    """returns an db value saying that this hash exists somewhere on a remote

    Returns
    -------
    bytes
        hash data db value
    """
    return f'50:{schema_hash}'.encode()


# ------------------------- Accessor Object -----------------------------------


class REMOTE_50_Options(metaclass=checkedmeta):
    _backend_options = EmptyDict()

    def __init__(self, backend_options, *args, **kwargs):
        if backend_options is None:
            backend_options = self.default_options
        self._backend_options = backend_options

    @property
    def default_options(self):
        return {}

    @property
    def backend_options(self):
        return self._backend_options

    @property
    def init_requires(self):
        return ('repo_path',)


class REMOTE_50_Handler(object):

    def __init__(self, repo_path: Path, *args, **kwargs):
        self.repo_path = repo_path
        self._dflt_backend_opts: Optional[dict] = None
        self._mode: Optional[str] = None

    def __enter__(self):
        return self

    def __exit__(self, *exc):
        return

    @reader_checkout_only
    def __getstate__(self) -> dict:  # pragma: no cover
        """ensure multiprocess operations can pickle relevant data.
        """
        self.close()
        state = self.__dict__.copy()
        return state

    def __setstate__(self, state: dict) -> None:  # pragma: no cover
        """ensure multiprocess operations can pickle relevant data.
        """
        self.__dict__.update(state)
        self.open(mode=self._mode)

    @property
    def backend_opts(self):
        return self._dflt_backend_opts

    @writer_checkout_only
    def _backend_opts_set(self, val):
        """Nonstandard descriptor method. See notes in ``backend_opts.setter``.
        """
        self._dflt_backend_opts = val
        return

    @backend_opts.setter
    def backend_opts(self, value):
        """
        Using seperate setter method (with ``@writer_checkout_only`` decorator
        applied) due to bug in python <3.8.

        From: https://bugs.python.org/issue19072
            > The classmethod decorator when applied to a function of a class,
            > does not honour the descriptor binding protocol for whatever it
            > wraps. This means it will fail when applied around a function which
            > has a decorator already applied to it and where that decorator
            > expects that the descriptor binding protocol is executed in order
            > to properly bind the function to the class.
        """
        return self._backend_opts_set(value)

    def open(self, mode, *args, **kwargs):
        self._mode = mode
        return

    def close(self, *args, **kwargs):
        return

    @staticmethod
    def delete_in_process_data(*args, **kwargs) -> None:
        """mockup of clearing staged directory for upstream calls.
        """
        return

    def read_data(self, hashVal: REMOTE_50_DataHashSpec) -> None:
        raise FileNotFoundError(
            f'data hash spec: {REMOTE_50_DataHashSpec} does not exist on this machine. '
            f'Perform a `data-fetch` operation to retrieve it from the remote server.')

    def write_data(self, schema_hash: str, *args, **kwargs) -> bytes:
        """Provide a formatted byte representation for storage as a remote reference

        Parameters
        ----------
        schema_hash : str
            schema hash which the referenced data sample should be accessed under

        Returns
        -------
        bytes
            formated raw values encoding lookup information
        """
        return remote_50_encode(schema_hash=schema_hash)




cdef class CompatibleData:
    """Bool recording if data `compatible` and if False the rejection `reason`.
    """

    def __init__(self, bint compatible, str reason):
        self.compatible = compatible
        self.reason = reason

    def __repr__(self):
        return (f'{self.__class__.__name__}('
                f'compatible={self.compatible}, '
                f'reason="{self.reason}")')

    def __iter__(self):
        for attr in ['compatible', 'reason']:
            yield getattr(self, attr)

    def __eq__(self, other):
        return (isinstance(other, self.__class__) and
                self.compatible == other.compatible and
                self.reason == other.reason)

    def __hash__(self):
        return hash((self.__class__, self.compatible, self.reason))


cdef class ColumnSchemaKey:
    """Record listing `column` name and `layout` type.
    """

    def __init__(self, str column, str layout):
        self.column = column
        self.layout = layout

    def __repr__(self):
        return (f'{self.__class__.__name__}('
                f'column="{self.column}", '
                f'layout="{self.layout}")')

    def __iter__(self):
        for attr in ['column', 'layout']:
            yield getattr(self, attr)

    def __eq__(self, other):
        return (isinstance(other, self.__class__) and
                self.column == other.column and
                self.layout == other.layout)

    def __hash__(self):
        return hash((self.__class__, self.column, self.layout))


cdef class FlatColumnDataKey:
    """Record listing `column` & `sample` name along with `layout` property
    """

    def __init__(self, str column, str sample):
        self.column = column
        self._sample = sample
        self._s_int = True if sample[0] == '#' else False

    def __repr__(self):
        return (f'{self.__class__.__name__}('
                f'column="{self.column}", '
                f'sample={f"{self.sample if self._s_int else repr(self.sample)}"})')

    def __iter__(self):
        for attr in ['column', 'sample']:
            yield getattr(self, attr)

    def __eq__(self, other):
        return (isinstance(other, self.__class__) and
                self.column == other.column and
                self.sample == other.sample)

    def __hash__(self):
        return hash((self.__class__, self.column, self.sample))

    @property
    def sample(self):
        if self._s_int:
            return int(self._sample[1:])
        else:
            return self._sample

    @property
    def layout(self):
        return 'flat'


cdef class NestedColumnDataKey:
    """Record listing `column`, `sample`, & `subsample` name along with `layout` property
    """

    def __init__(self, str column, str sample, str subsample):
        self.column = column
        self._sample = sample
        self._subsample = subsample
        self._s_int = True if sample[0] == '#' else False
        self._ss_int = True if subsample[0] == '#' else False

    def __repr__(self):
        return (f'{self.__class__.__name__}('
                f'column="{self.column}", '
                f'sample={f"{self.sample if self._s_int else repr(self.sample)}"}, '
                f'subsample={f"{self.subsample if self._ss_int else repr(self.subsample)}"})')

    def __iter__(self):
        for attr in ['column', 'sample', 'subsample']:
            yield getattr(self, attr)

    def __eq__(self, other):
        return (isinstance(other, self.__class__) and
                self.column == other.column and
                self.sample == other.sample and
                self.subsample == other.subsample)

    def __hash__(self):
        return hash((self.__class__, self.column, self.sample, self.subsample))

    @property
    def sample(self):
        if self._s_int:
            return int(self._sample[1:])
        else:
            return self._sample

    @property
    def subsample(self):
        if self._ss_int:
            return int(self._subsample[1:])
        else:
            return self._subsample

    @property
    def layout(self):
        return 'nested'


cdef class DataRecordVal:

    def __init__(self, str digest):
        self.digest = digest

    def __repr__(self):
        return (f'{self.__class__.__name__}('
                f'digest={repr(self.digest)})')

    def __iter__(self):
        for attr in ['digest']:
            yield getattr(self, attr)

    def __eq__(self, other):
        return (isinstance(other, self.__class__)
                and self.digest == other.digest)

    def __hash__(self):
        return hash((self.__class__, self.digest))



import pytest
import numpy as np


from hangar.typesystem import StringVariableShape


class TestInvalidValues:

    @pytest.mark.parametrize('coltype', ['ndarray', np.ndarray, 32, {'foo': 'bar'}, ascii])
    def test_column_type_must_be_str(self, coltype):
        with pytest.raises(ValueError):
            StringVariableShape(dtype=str, column_layout='flat', column_type=coltype)

    @pytest.mark.parametrize('collayout', ['f', 'n', None, 32, {'foo': 'bar'}, ascii])
    def test_column_layout_must_be_valid_value(self, collayout):
        with pytest.raises(ValueError):
            StringVariableShape(dtype=str, column_layout=collayout)

    @pytest.mark.parametrize('backend', ['00', 24, {'30': '30'}, ('30',), ['50',], ascii, 'None'])
    def test_variable_shape_backend_code_valid_value(self, backend):
        with pytest.raises(ValueError):
            StringVariableShape(dtype=str, column_layout='flat', backend=backend)

    @pytest.mark.parametrize('opts', ['val', [], (), [('k', 'v')], 10, ({'k': 'v'},), ascii])
    def test_backend_options_must_be_dict_or_nonetype(self, opts):
        with pytest.raises(TypeError):
            StringVariableShape(dtype=str, column_layout='flat', backend='30', backend_options=opts)

    def test_backend_must_be_specified_if_backend_options_provided(self):
        with pytest.raises(ValueError):
            StringVariableShape(dtype=str, column_layout='flat', backend_options={})

    @pytest.mark.parametrize('schema_type', ['fixed_shape', True, 'str', np.uint8, 3, ascii])
    def test_variable_shape_must_have_variable_shape_schema_type(self, schema_type):
        with pytest.raises(ValueError):
            StringVariableShape(dtype=str, column_layout='flat', schema_type=schema_type)


